# audiotoolbox.pxd
# AudioToolbox framework declarations for coremusic

from .coreaudio cimport *
from .coreaudiotypes cimport *
from .corefoundation cimport *

# -----------------------------------------------------------------------------

cdef extern from "AudioToolbox/AudioQueue.h":

    ctypedef UInt32 AudioQueuePropertyID
    ctypedef UInt32 AudioQueueParameterID
    ctypedef Float32 AudioQueueParameterValue
    ctypedef struct OpaqueAudioQueue
    ctypedef OpaqueAudioQueue* AudioQueueRef
    ctypedef struct OpaqueAudioQueueTimeline
    ctypedef OpaqueAudioQueueTimeline* AudioQueueTimelineRef

    ctypedef struct AudioQueueBuffer:
        UInt32 mAudioDataBytesCapacity
        void* mAudioData
        UInt32 mAudioDataByteSize
        void* mUserData
        UInt32 mPacketDescriptionCapacity
        # AudioStreamPacketDescription* mPacketDescriptions
        UInt32 mPacketDescriptionCount

    ctypedef AudioQueueBuffer* AudioQueueBufferRef

    ctypedef void (*AudioQueueOutputCallback)(void* inUserData, AudioQueueRef inAQ, AudioQueueBufferRef inBuffer)
    ctypedef void (*AudioQueueInputCallback)(void* inUserData, AudioQueueRef inAQ, AudioQueueBufferRef inBuffer, const AudioTimeStamp* inStartTime, UInt32 inNumberPacketDescriptions, const AudioStreamPacketDescription* inPacketDescs)

    ctypedef enum:
        kAudioQueueErr_InvalidBuffer = -66687
        kAudioQueueErr_BufferEmpty = -66686
        kAudioQueueErr_DisposalPending = -66685
        kAudioQueueErr_InvalidProperty = -66684
        kAudioQueueErr_InvalidPropertySize = -66683
        kAudioQueueErr_InvalidParameter = -66682
        kAudioQueueErr_CannotStart = -66681
        kAudioQueueErr_InvalidDevice = -66680
        kAudioQueueErr_BufferInQueue = -66679
        kAudioQueueErr_InvalidRunState = -66678
        kAudioQueueErr_InvalidQueueType = -66677
        kAudioQueueErr_Permissions = -66676
        kAudioQueueErr_InvalidPropertyValue = -66675
        kAudioQueueErr_PrimeTimedOut = -66674
        kAudioQueueErr_CodecNotFound = -66673
        kAudioQueueErr_InvalidCodecAccess = -66672
        kAudioQueueErr_QueueInvalidated = -66671
        kAudioQueueErr_TooManyTaps = -66670
        kAudioQueueErr_InvalidTapContext = -66669
        kAudioQueueErr_RecordUnderrun = -66668
        kAudioQueueErr_InvalidTapType = -66667
        kAudioQueueErr_BufferEnqueuedTwice = -66666
        kAudioQueueErr_CannotStartYet = -66665
        kAudioQueueErr_EnqueueDuringReset = -66632
        kAudioQueueErr_InvalidOfflineMode = -66626

    cdef OSStatus AudioQueueNewOutput(const AudioStreamBasicDescription* inFormat, AudioQueueOutputCallback inCallbackProc, void* inUserData, CFRunLoopRef inCallbackRunLoop, CFStringRef inCallbackRunLoopMode, UInt32 inFlags, AudioQueueRef* outAQ)
    cdef OSStatus AudioQueueNewInput(const AudioStreamBasicDescription* inFormat, AudioQueueInputCallback inCallbackProc, void* inUserData, CFRunLoopRef inCallbackRunLoop, CFStringRef inCallbackRunLoopMode, UInt32 inFlags, AudioQueueRef* outAQ)
    cdef OSStatus AudioQueueDispose(AudioQueueRef inAQ, Boolean inImmediate)
    cdef OSStatus AudioQueueAllocateBuffer(AudioQueueRef inAQ, UInt32 inBufferByteSize, AudioQueueBufferRef* outBuffer)
    cdef OSStatus AudioQueueAllocateBufferWithPacketDescriptions(AudioQueueRef inAQ, UInt32 inBufferByteSize, UInt32 inNumberPacketDescriptions, AudioQueueBufferRef* outBuffer)
    cdef OSStatus AudioQueueFreeBuffer(AudioQueueRef inAQ, AudioQueueBufferRef inBuffer)
    cdef OSStatus AudioQueueEnqueueBuffer(AudioQueueRef inAQ, AudioQueueBufferRef inBuffer, UInt32 inNumPacketDescs, const void* inPacketDescs) nogil
    cdef OSStatus AudioQueueEnqueueBufferWithParameters(AudioQueueRef inAQ, AudioQueueBufferRef inBuffer, UInt32 inNumPacketDescs, const void* inPacketDescs, UInt32 inTrimFramesAtStart, UInt32 inTrimFramesAtEnd, UInt32 inNumParamValues, const AudioQueueParameterEvent* inParamValues, const AudioTimeStamp* inStartTime, AudioTimeStamp* outActualStartTime) nogil
    cdef OSStatus AudioQueueStart(AudioQueueRef inAQ, const AudioTimeStamp* inStartTime)
    cdef OSStatus AudioQueuePrime(AudioQueueRef inAQ, UInt32 inNumberOfFramesToPrepare, UInt32* outNumberOfFramesPrepared)
    cdef OSStatus AudioQueueStop(AudioQueueRef inAQ, Boolean inImmediate)
    cdef OSStatus AudioQueuePause(AudioQueueRef inAQ)
    cdef OSStatus AudioQueueFlush(AudioQueueRef inAQ)
    cdef OSStatus AudioQueueReset(AudioQueueRef inAQ)
    cdef OSStatus AudioQueueGetCurrentTime(AudioQueueRef inAQ, AudioQueueTimelineRef inTimeline, AudioTimeStamp* outTimeStamp, Boolean* outTimelineDiscontinuity)


cdef extern from "AudioToolbox/AudioFile.h":

    ctypedef UInt32 AudioFileTypeID
    ctypedef UInt32 AudioFilePropertyID
    ctypedef UInt32 AudioFileFlags
    ctypedef struct OpaqueAudioFileID
    ctypedef OpaqueAudioFileID* AudioFileID

    # AudioFileFlags constants
    ctypedef enum:
        kAudioFileFlags_EraseFile = 1
        kAudioFileFlags_DontPageAlignAudioData = 2

    # Audio file type IDs
    ctypedef enum:
        kAudioFileAIFFType = 1095321158  # 'AIFF'
        kAudioFileAIFCType = 1095321155  # 'AIFC'
        kAudioFileWAVEType = 1463899717  # 'WAVE'
        kAudioFileRF64Type = 1380333108  # 'RF64'
        kAudioFileBW64Type = 1112493108  # 'BW64'
        kAudioFileWave64Type = 1463900518  # 'W64f'
        kAudioFileSoundDesigner2Type = 1399075686  # 'Sd2f'
        kAudioFileNextType = 1315264596  # 'NeXT'
        kAudioFileMP3Type = 1297106739  # 'MPG3'
        kAudioFileMP2Type = 1297106738  # 'MPG2'
        kAudioFileMP1Type = 1297106737  # 'MPG1'
        kAudioFileAC3Type = 1633889587  # 'ac-3'
        kAudioFileAAC_ADTSType = 1633973363  # 'adts'
        kAudioFileMPEG4Type = 1836069990  # 'mp4f'
        kAudioFileM4AType = 1832149350  # 'm4af'
        kAudioFileM4BType = 1832149606  # 'm4bf'
        kAudioFileCAFType = 1667327590  # 'caff'
        kAudioFile3GPType = 862417008   # '3gpp'
        kAudioFile3GP2Type = 862416690  # '3gp2'
        kAudioFileAMRType = 1634562662  # 'amrf'
        kAudioFileFLACType = 1718378851 # 'flac'
        kAudioFileLATMInLOASType = 1819238259 # 'loas'

    ctypedef enum:
        kAudioFileReadPermission = 1
        kAudioFileWritePermission = 2
        kAudioFileReadWritePermission = 3

    ctypedef enum:
        kAudioFilePropertyFileFormat = 1717988724         # 'ffmt'
        kAudioFilePropertyDataFormat = 1684103783         # 'dfmt'
        kAudioFilePropertyIsOptimized = 1869640813        # 'optm'
        kAudioFilePropertyMagicCookieData = 1835493731    # 'mgic'
        kAudioFilePropertyAudioDataByteCount = 1650683508 # 'bcnt'
        kAudioFilePropertyAudioDataPacketCount = 1885564532 # 'pcnt'
        kAudioFilePropertyMaximumPacketSize = 1886616691  # 'psze'
        kAudioFilePropertyDataOffset = 1685022310         # 'doff'
        kAudioFilePropertyChannelLayout = 1668112500      # 'cmap'
        kAudioFilePropertyDeferSizeUpdates = 1684238953   # 'dszu'
        kAudioFilePropertyDataFormatName = 1718512996     # 'fnme'
        kAudioFilePropertyMarkerList = 1835756403         # 'mkls'
        kAudioFilePropertyRegionList = 1919380595         # 'rgls'
        kAudioFilePropertyPacketToFrame = 1886086770      # 'pkfr'
        kAudioFilePropertyFrameToPacket = 1718775151      # 'frpk'
        kAudioFilePropertyPacketToByte = 1886085753       # 'pkby'
        kAudioFilePropertyByteToPacket = 1652125803       # 'bypk'
        kAudioFilePropertyChunkIDs = 1667787108           # 'chid'
        kAudioFilePropertyInfoDictionary = 1768842863     # 'info'
        kAudioFilePropertyPacketTableInfo = 1886283375    # 'pnfo'
        kAudioFilePropertyFormatList = 1718383476         # 'flst'
        kAudioFilePropertyPacketSizeUpperBound = 1886090093 # 'pkub'
        kAudioFilePropertyReserveDuration = 1920365423    # 'rsrv'
        kAudioFilePropertyEstimatedDuration = 1701082482  # 'edur'
        kAudioFilePropertyBitRate = 1651663220            # 'brat'
        kAudioFilePropertyID3Tag = 1768174180             # 'id3 '
        kAudioFilePropertySourceBitDepth = 1935832164     # 'sbtd'
        kAudioFilePropertyAlbumArtwork = 1635015020       # 'aart'
        kAudioFilePropertyAudioTrackCount = 1635017588    # 'atct'
        kAudioFilePropertyUseAudioTrack = 1969385580      # 'uatk'

    cdef OSStatus AudioFileOpenURL(CFURLRef inFileRef, AudioFilePermissions inPermissions, AudioFileTypeID inFileTypeHint, AudioFileID* outAudioFile)
    cdef OSStatus AudioFileCreateWithURL(CFURLRef inFileRef, AudioFileTypeID inFileType, const AudioStreamBasicDescription* inFormat, AudioFileFlags inFlags, AudioFileID* outAudioFile)
    cdef OSStatus AudioFileClose(AudioFileID inAudioFile)
    cdef OSStatus AudioFileOptimize(AudioFileID inAudioFile)
    cdef OSStatus AudioFileReadBytes(AudioFileID inAudioFile, Boolean inUseCache, SInt64 inStartingByte, UInt32* ioNumBytes, void* outBuffer)
    cdef OSStatus AudioFileWriteBytes(AudioFileID inAudioFile, Boolean inUseCache, SInt64 inStartingByte, UInt32* ioNumBytes, const void* inBuffer)
    cdef OSStatus AudioFileReadPackets(AudioFileID inAudioFile, Boolean inUseCache, UInt32* ioNumBytes, void* outPacketDescriptions, SInt64 inStartingPacket, UInt32* ioNumPackets, void* outBuffer)
    cdef OSStatus AudioFileReadPacketData(AudioFileID inAudioFile, Boolean inUseCache, UInt32* ioNumBytes, void* outPacketDescriptions, SInt64 inStartingPacket, UInt32* ioNumPackets, void* outBuffer)
    cdef OSStatus AudioFileWritePackets(AudioFileID inAudioFile, Boolean inUseCache, UInt32 inNumBytes, const void* inPacketDescriptions, SInt64 inStartingPacket, UInt32* ioNumPackets, const void* inBuffer)
    cdef OSStatus AudioFileCountUserData(AudioFileID inAudioFile, UInt32 inUserDataID, UInt32* outNumberItems)
    cdef OSStatus AudioFileGetUserDataSize(AudioFileID inAudioFile, UInt32 inUserDataID, UInt32 inIndex, UInt32* outUserDataSize)
    cdef OSStatus AudioFileGetUserData(AudioFileID inAudioFile, UInt32 inUserDataID, UInt32 inIndex, UInt32* ioUserDataSize, void* outUserData)
    cdef OSStatus AudioFileSetUserData(AudioFileID inAudioFile, UInt32 inUserDataID, UInt32 inIndex, UInt32 inUserDataSize, const void* inUserData)
    cdef OSStatus AudioFileRemoveUserData(AudioFileID inAudioFile, UInt32 inUserDataID, UInt32 inIndex)
    cdef OSStatus AudioFileGetPropertyInfo(AudioFileID inAudioFile, AudioFilePropertyID inPropertyID, UInt32* outDataSize, UInt32* isWritable)
    cdef OSStatus AudioFileGetProperty(AudioFileID inAudioFile, AudioFilePropertyID inPropertyID, UInt32* ioDataSize, void* outPropertyData)
    cdef OSStatus AudioFileSetProperty(AudioFileID inAudioFile, AudioFilePropertyID inPropertyID, UInt32 inDataSize, const void* inPropertyData)

    ctypedef struct AudioQueueParameterEvent:
        AudioQueueParameterID mID
        AudioQueueParameterValue mValue


cdef extern from "AudioToolbox/AudioFileStream.h":

    # AudioFileStream types
    ctypedef UInt32 AudioFileStreamPropertyID
    ctypedef struct OpaqueAudioFileStreamID
    ctypedef OpaqueAudioFileStreamID* AudioFileStreamID

    # AudioFileStream flags
    ctypedef enum AudioFileStreamPropertyFlags:
        kAudioFileStreamPropertyFlag_PropertyIsCached = 1
        kAudioFileStreamPropertyFlag_CacheProperty = 2

    ctypedef enum AudioFileStreamParseFlags:
        kAudioFileStreamParseFlag_Discontinuity = 1

    ctypedef enum AudioFileStreamSeekFlags:
        kAudioFileStreamSeekFlag_OffsetIsEstimated = 1

    # AudioFileStream callback function types
    ctypedef void (*AudioFileStream_PropertyListenerProc)(
        void* inClientData,
        AudioFileStreamID inAudioFileStream,
        AudioFileStreamPropertyID inPropertyID,
        AudioFileStreamPropertyFlags* ioFlags)

    ctypedef void (*AudioFileStream_PacketsProc)(
        void* inClientData,
        UInt32 inNumberBytes,
        UInt32 inNumberPackets,
        const void* inInputData,
        AudioStreamPacketDescription* inPacketDescriptions)

    # AudioFileStream error codes
    ctypedef enum:
        kAudioFileStreamError_UnsupportedFileType = 1953064820  # 'typ?'
        kAudioFileStreamError_UnsupportedDataFormat = 1718449215  # 'fmt?'
        kAudioFileStreamError_UnsupportedProperty = 1886547839  # 'pty?'
        kAudioFileStreamError_BadPropertySize = 561211770  # '!siz'
        kAudioFileStreamError_NotOptimized = 1869640813  # 'optm'
        kAudioFileStreamError_InvalidPacketOffset = 1885433391  # 'pck?'
        kAudioFileStreamError_InvalidFile = 1685348415  # 'dta?'
        kAudioFileStreamError_ValueUnknown = 1970170687  # 'unk?'
        kAudioFileStreamError_DataUnavailable = 1836016741  # 'more'
        kAudioFileStreamError_IllegalOperation = 1852139888  # 'nope'
        kAudioFileStreamError_UnspecifiedError = 2003395684  # 'wht?'
        kAudioFileStreamError_DiscontinuityCantRecover = 1685348641  # 'dsc!'

    # AudioFileStream property IDs
    ctypedef enum:
        kAudioFileStreamProperty_ReadyToProducePackets = 1919247473  # 'redy'
        kAudioFileStreamProperty_FileFormat = 1717988724  # 'ffmt'
        kAudioFileStreamProperty_DataFormat = 1684103783  # 'dfmt'
        kAudioFileStreamProperty_FormatList = 1718383476  # 'flst'
        kAudioFileStreamProperty_MagicCookieData = 1835493731  # 'mgic'
        kAudioFileStreamProperty_AudioDataByteCount = 1650683508  # 'bcnt'
        kAudioFileStreamProperty_AudioDataPacketCount = 1885564532  # 'pcnt'
        kAudioFileStreamProperty_MaximumPacketSize = 1886616691  # 'psze'
        kAudioFileStreamProperty_DataOffset = 1685022310  # 'doff'
        kAudioFileStreamProperty_ChannelLayout = 1668112500  # 'cmap'
        kAudioFileStreamProperty_PacketToFrame = 1886086770  # 'pkfr'
        kAudioFileStreamProperty_FrameToPacket = 1718775151  # 'frpk'
        kAudioFileStreamProperty_RestrictsRandomAccess = 1919508592  # 'rrap'
        kAudioFileStreamProperty_PacketToRollDistance = 1886090604  # 'pkrl'
        kAudioFileStreamProperty_PreviousIndependentPacket = 1885957228  # 'pind'
        kAudioFileStreamProperty_NextIndependentPacket = 1852273252  # 'nind'
        kAudioFileStreamProperty_PacketToDependencyInfo = 1886086256  # 'pkdp'
        kAudioFileStreamProperty_PacketToByte = 1886085753  # 'pkby'
        kAudioFileStreamProperty_ByteToPacket = 1652125803  # 'bypk'
        kAudioFileStreamProperty_PacketTableInfo = 1886283375  # 'pnfo'
        kAudioFileStreamProperty_PacketSizeUpperBound = 1886090093  # 'pkub'
        kAudioFileStreamProperty_AverageBytesPerPacket = 1633969264  # 'abpp'
        kAudioFileStreamProperty_BitRate = 1651663220  # 'brat'
        kAudioFileStreamProperty_InfoDictionary = 1768842863  # 'info'

    # AudioFileStream functions
    cdef OSStatus AudioFileStreamOpen(
        void* inClientData,
        AudioFileStream_PropertyListenerProc inPropertyListenerProc,
        AudioFileStream_PacketsProc inPacketsProc,
        AudioFileTypeID inFileTypeHint,
        AudioFileStreamID* outAudioFileStream)

    cdef OSStatus AudioFileStreamParseBytes(
        AudioFileStreamID inAudioFileStream,
        UInt32 inDataByteSize,
        const void* inData,
        AudioFileStreamParseFlags inFlags)

    cdef OSStatus AudioFileStreamSeek(
        AudioFileStreamID inAudioFileStream,
        SInt64 inPacketOffset,
        SInt64* outDataByteOffset,
        AudioFileStreamSeekFlags* ioFlags)

    cdef OSStatus AudioFileStreamGetPropertyInfo(
        AudioFileStreamID inAudioFileStream,
        AudioFileStreamPropertyID inPropertyID,
        UInt32* outPropertyDataSize,
        Boolean* outWritable)

    cdef OSStatus AudioFileStreamGetProperty(
        AudioFileStreamID inAudioFileStream,
        AudioFileStreamPropertyID inPropertyID,
        UInt32* ioPropertyDataSize,
        void* outPropertyData)

    cdef OSStatus AudioFileStreamSetProperty(
        AudioFileStreamID inAudioFileStream,
        AudioFileStreamPropertyID inPropertyID,
        UInt32 inPropertyDataSize,
        const void* inPropertyData)

    cdef OSStatus AudioFileStreamClose(AudioFileStreamID inAudioFileStream)


cdef extern from "AudioToolbox/AudioComponent.h":

    ctypedef struct OpaqueAudioComponent
    ctypedef OpaqueAudioComponent* AudioComponent
    ctypedef struct OpaqueAudioComponentInstance
    ctypedef OpaqueAudioComponentInstance* AudioComponentInstance

    ctypedef struct AudioComponentDescription:
        UInt32 componentType
        UInt32 componentSubType
        UInt32 componentManufacturer
        UInt32 componentFlags
        UInt32 componentFlagsMask

    ctypedef OSStatus (*AudioComponentFactoryFunction)(AudioComponentDescription* inDescription, AudioComponent inComponent, AudioComponentInstance* outInstance)
    ctypedef void (*AudioComponentInitializer)(AudioComponentInstance inInstance)

    # AudioComponent constants
    ctypedef enum:
        kAudioComponentType_Output = 1635086197          # 'auou'
        kAudioComponentType_MusicDevice = 1635085685     # 'aumu'
        kAudioComponentType_MusicEffect = 1635085670     # 'aumf'
        kAudioComponentType_FormatConverter = 1635083875 # 'aufc'
        kAudioComponentType_Effect = 1635083896          # 'aufx'
        kAudioComponentType_Mixer = 1635085688           # 'aumx'
        kAudioComponentType_Panner = 1635086446          # 'aupn'
        kAudioComponentType_Generator = 1635084142       # 'augn'
        kAudioComponentType_OfflineEffect = 1635085676   # 'auol'
        kAudioComponentType_MIDIProcessor = 1635085673   # 'aumi'

        kAudioComponentSubType_DefaultOutput = 1684366880        # 'def '
        kAudioComponentSubType_HALOutput = 1634230896            # 'ahal'
        kAudioComponentSubType_SystemOutput = 1937339252         # 'sys '
        kAudioComponentSubType_GenericOutput = 1734700658       # 'genr'

        kAudioComponentManufacturer_Apple = 1634758764  # 'appl'

    # AudioComponent functions
    cdef AudioComponent AudioComponentFindNext(AudioComponent inComponent, const AudioComponentDescription* inDesc)
    cdef OSStatus AudioComponentCopyName(AudioComponent inComponent, CFStringRef* outName)
    cdef OSStatus AudioComponentGetDescription(AudioComponent inComponent, AudioComponentDescription* outDesc)
    cdef OSStatus AudioComponentGetVersion(AudioComponent inComponent, UInt32* outVersion)
    cdef OSStatus AudioComponentInstanceNew(AudioComponent inComponent, AudioComponentInstance* outInstance)
    cdef OSStatus AudioComponentInstanceDispose(AudioComponentInstance inInstance)
    cdef OSStatus AudioComponentInstanceCanDo(AudioComponentInstance inInstance, SInt16 inSelectorID)


cdef extern from "AudioToolbox/AudioUnitProperties.h":
    pass


cdef extern from "AudioToolbox/AUComponent.h":

    # AudioUnit is typedef'd as AudioComponentInstance
    ctypedef AudioComponentInstance AudioUnit

    # AudioUnit Types (same as AudioComponent types but with different name)
    ctypedef enum:
        kAudioUnitType_Output = 1635086197               # 'auou'
        kAudioUnitType_MusicDevice = 1635085685          # 'aumu'
        kAudioUnitType_MusicEffect = 1635085670          # 'aumf'
        kAudioUnitType_FormatConverter = 1635083875      # 'aufc'
        kAudioUnitType_Effect = 1635083896               # 'aufx'
        kAudioUnitType_Mixer = 1635085688                # 'aumx'
        kAudioUnitType_Panner = 1635086446               # 'aupn'
        kAudioUnitType_Generator = 1635084142            # 'augn'
        kAudioUnitType_OfflineEffect = 1635085676        # 'auol'
        kAudioUnitType_MIDIProcessor = 1635085673        # 'aumi'

        kAudioUnitSubType_DefaultOutput = 1684366880     # 'def '
        kAudioUnitSubType_HALOutput = 1634230896         # 'ahal'
        kAudioUnitSubType_SystemOutput = 1937339252      # 'sys '
        kAudioUnitSubType_GenericOutput = 1734700658     # 'genr'

        kAudioUnitManufacturer_Apple = 1634758764        # 'appl'

    # AudioUnit Property IDs
    ctypedef UInt32 AudioUnitPropertyID
    ctypedef UInt32 AudioUnitScope
    ctypedef UInt32 AudioUnitElement
    ctypedef UInt32 AudioUnitParameterID
    ctypedef Float32 AudioUnitParameterValue

    ctypedef enum:
        # Global scope properties
        kAudioUnitProperty_ClassInfo = 0
        kAudioUnitProperty_MakeConnection = 1
        kAudioUnitProperty_SampleRate = 2
        kAudioUnitProperty_ParameterList = 3
        kAudioUnitProperty_ParameterInfo = 4
        kAudioUnitProperty_CPULoad = 6
        kAudioUnitProperty_StreamFormat = 8
        kAudioUnitProperty_ElementCount = 11
        kAudioUnitProperty_Latency = 12
        kAudioUnitProperty_SupportedNumChannels = 13
        kAudioUnitProperty_MaximumFramesPerSlice = 14
        kAudioUnitProperty_ParameterValueStrings = 16
        kAudioUnitProperty_AudioChannelLayout = 19
        kAudioUnitProperty_TailTime = 20
        kAudioUnitProperty_BypassEffect = 21
        kAudioUnitProperty_LastRenderError = 22
        kAudioUnitProperty_SetRenderCallback = 23
        kAudioUnitProperty_FactoryPresets = 24
        kAudioUnitProperty_RenderQuality = 26
        kAudioUnitProperty_HostCallbacks = 27
        kAudioUnitProperty_InPlaceProcessing = 29
        kAudioUnitProperty_ElementName = 30
        kAudioUnitProperty_SupportedChannelLayoutTags = 32
        kAudioUnitProperty_PresentPreset = 36
        kAudioUnitProperty_OfflineRender = 37
        kAudioUnitProperty_ParameterValueName = 38
        kAudioUnitProperty_ParameterStringFromValue = 39
        kAudioUnitProperty_ParameterValueFromString = 40
        kAudioUnitProperty_NickName = 54

    # AudioUnit Parameter Units
    ctypedef UInt32 AudioUnitParameterUnit
    ctypedef enum:
        kAudioUnitParameterUnit_Generic = 0
        kAudioUnitParameterUnit_Indexed = 1
        kAudioUnitParameterUnit_Boolean = 2
        kAudioUnitParameterUnit_Percent = 3
        kAudioUnitParameterUnit_Seconds = 4
        kAudioUnitParameterUnit_SampleFrames = 5
        kAudioUnitParameterUnit_Phase = 6
        kAudioUnitParameterUnit_Rate = 7
        kAudioUnitParameterUnit_Hertz = 8
        kAudioUnitParameterUnit_Cents = 9
        kAudioUnitParameterUnit_RelativeSemiTones = 10
        kAudioUnitParameterUnit_MIDINoteNumber = 11
        kAudioUnitParameterUnit_MIDIController = 12
        kAudioUnitParameterUnit_Decibels = 13
        kAudioUnitParameterUnit_LinearGain = 14
        kAudioUnitParameterUnit_Degrees = 15
        kAudioUnitParameterUnit_EqualPowerCrossfade = 16
        kAudioUnitParameterUnit_MixerFaderCurve1 = 17
        kAudioUnitParameterUnit_Pan = 18
        kAudioUnitParameterUnit_Meters = 19
        kAudioUnitParameterUnit_AbsoluteCents = 20
        kAudioUnitParameterUnit_Octaves = 21
        kAudioUnitParameterUnit_BPM = 22
        kAudioUnitParameterUnit_Beats = 23
        kAudioUnitParameterUnit_Milliseconds = 24
        kAudioUnitParameterUnit_Ratio = 25
        kAudioUnitParameterUnit_CustomUnit = 26

    # AudioUnit Parameter Flags
    ctypedef UInt32 AudioUnitParameterFlags
    ctypedef enum:
        kAudioUnitParameterFlag_CFNameRelease = (1 << 4)
        kAudioUnitParameterFlag_HasClump = (1 << 20)
        kAudioUnitParameterFlag_HasName = (1 << 21)
        kAudioUnitParameterFlag_DisplayLogarithmic = (1 << 22)
        kAudioUnitParameterFlag_DisplaySquareRoot = (1 << 23)
        kAudioUnitParameterFlag_DisplaySquared = (1 << 24)
        kAudioUnitParameterFlag_DisplayCubed = (1 << 25)
        kAudioUnitParameterFlag_DisplayCubeRoot = (1 << 26)
        kAudioUnitParameterFlag_DisplayExponential = (1 << 27)
        kAudioUnitParameterFlag_HasCFNameString = (1 << 28)
        kAudioUnitParameterFlag_IsGlobalMeta = (1 << 29)
        kAudioUnitParameterFlag_IsElementMeta = (1 << 30)
        kAudioUnitParameterFlag_IsReadable = (1 << 31)
        kAudioUnitParameterFlag_IsWritable = (1 << 31) >> 1

    # AudioUnit Parameter Info structure
    ctypedef struct AudioUnitParameterInfo:
        char name[52]
        CFStringRef unitName
        UInt32 clumpID
        CFStringRef cfNameString
        AudioUnitParameterUnit unit
        AudioUnitParameterValue minValue
        AudioUnitParameterValue maxValue
        AudioUnitParameterValue defaultValue
        UInt32 flags

    # AudioUnit Preset (AUPreset) structure
    ctypedef struct AUPreset:
        SInt32 presetNumber
        CFStringRef presetName

    # AudioUnit Preset dictionary keys (for kAudioUnitProperty_FactoryPresets)
    # These are #define macros, not CFStringRef constants
    # #define kAUPresetNumberKey "preset-number"
    # #define kAUPresetNameKey "name"

    # AudioUnit Parameter Value translation structure
    ctypedef struct AudioUnitParameterValueTranslation:
        AudioUnit inUnit
        AudioUnitParameterID inParameterID
        AudioUnitScope inScope
        AudioUnitElement inElement
        CFStringRef inString
        AudioUnitParameterValue outValue

    # AudioUnit Scopes
    ctypedef enum:
        kAudioUnitScope_Global = 0
        kAudioUnitScope_Input = 1
        kAudioUnitScope_Output = 2
        kAudioUnitScope_Group = 3
        kAudioUnitScope_Part = 4
        kAudioUnitScope_Note = 5
        kAudioUnitScope_Layer = 6
        kAudioUnitScope_LayerItem = 7

    # Render callback function type
    ctypedef UInt32 AudioUnitRenderActionFlags
    ctypedef OSStatus (*AURenderCallback)(void* inRefCon,
                                           AudioUnitRenderActionFlags* ioActionFlags,
                                           const AudioTimeStamp* inTimeStamp,
                                           UInt32 inBusNumber,
                                           UInt32 inNumberFrames,
                                           AudioBufferList* ioData)

    ctypedef struct AURenderCallbackStruct:
        AURenderCallback inputProc
        void* inputProcRefCon

    # AudioUnit functions
    cdef OSStatus AudioUnitInitialize(AudioUnit inUnit)
    cdef OSStatus AudioUnitUninitialize(AudioUnit inUnit)
    cdef OSStatus AudioUnitGetPropertyInfo(AudioUnit inUnit, AudioUnitPropertyID inID, AudioUnitScope inScope, AudioUnitElement inElement, UInt32* outDataSize, Boolean* outWritable)
    cdef OSStatus AudioUnitGetProperty(AudioUnit inUnit, AudioUnitPropertyID inID, AudioUnitScope inScope, AudioUnitElement inElement, void* outData, UInt32* ioDataSize)
    cdef OSStatus AudioUnitSetProperty(AudioUnit inUnit, AudioUnitPropertyID inID, AudioUnitScope inScope, AudioUnitElement inElement, const void* inData, UInt32 inDataSize)
    cdef OSStatus AudioUnitAddPropertyListener(AudioUnit inUnit, AudioUnitPropertyID inID, void* inProc, void* inProcUserData)
    cdef OSStatus AudioUnitRemovePropertyListener(AudioUnit inUnit, AudioUnitPropertyID inID, void* inProc)
    cdef OSStatus AudioUnitRemovePropertyListenerWithUserData(AudioUnit inUnit, AudioUnitPropertyID inID, void* inProc, void* inProcUserData)
    cdef OSStatus AudioUnitAddRenderNotify(AudioUnit inUnit, AURenderCallback inProc, void* inProcUserData)
    cdef OSStatus AudioUnitRemoveRenderNotify(AudioUnit inUnit, AURenderCallback inProc, void* inProcUserData)
    cdef OSStatus AudioUnitGetParameter(AudioUnit inUnit, AudioUnitParameterID inID, AudioUnitScope inScope, AudioUnitElement inElement, AudioUnitParameterValue* outValue)
    cdef OSStatus AudioUnitSetParameter(AudioUnit inUnit, AudioUnitParameterID inID, AudioUnitScope inScope, AudioUnitElement inElement, AudioUnitParameterValue inValue, UInt32 inBufferOffsetInFrames)
    cdef OSStatus AudioUnitScheduleParameters(AudioUnit inUnit, const void* inParameterEvent, UInt32 inNumParamEvents)
    cdef OSStatus AudioUnitRender(AudioUnit inUnit, AudioUnitRenderActionFlags* ioActionFlags, const AudioTimeStamp* inTimeStamp, UInt32 inOutputBusNumber, UInt32 inNumberFrames, AudioBufferList* ioData)
    cdef OSStatus AudioUnitProcess(AudioUnit inUnit, AudioUnitRenderActionFlags* ioActionFlags, const AudioTimeStamp* inTimeStamp, UInt32 inNumberFrames, AudioBufferList* ioData)
    cdef OSStatus AudioUnitProcessMultiple(AudioUnit inUnit, AudioUnitRenderActionFlags* ioActionFlags, const AudioTimeStamp* inTimeStamp, UInt32 inNumberFrames, UInt32 inNumberInputBufferLists, const AudioBufferList** inInputBufferLists, UInt32 inNumberOutputBufferLists, AudioBufferList** ioOutputBufferLists)
    cdef OSStatus AudioUnitReset(AudioUnit inUnit, AudioUnitScope inScope, AudioUnitElement inElement)


cdef extern from "AudioToolbox/AudioOutputUnit.h":

    cdef OSStatus AudioOutputUnitStart(AudioUnit ci)
    cdef OSStatus AudioOutputUnitStop(AudioUnit ci)


cdef extern from "AudioToolbox/AudioConverter.h":

    # AudioConverter types
    ctypedef struct OpaqueAudioConverter
    ctypedef OpaqueAudioConverter* AudioConverterRef
    ctypedef UInt32 AudioConverterPropertyID

    ctypedef enum AudioConverterOptions:
        kAudioConverterOption_Unbuffered = 65536

    # AudioConverter property IDs
    ctypedef enum:
        kAudioConverterPropertyMinimumInputBufferSize = 1835623027  # 'mibs'
        kAudioConverterPropertyMinimumOutputBufferSize = 1836016243  # 'mobs'
        kAudioConverterPropertyMaximumInputPacketSize = 2020175987  # 'xips'
        kAudioConverterPropertyMaximumOutputPacketSize = 2020569203  # 'xops'
        kAudioConverterPropertyCalculateInputBufferSize = 1667850867  # 'cibs'
        kAudioConverterPropertyCalculateOutputBufferSize = 1668244083  # 'cobs'
        kAudioConverterPropertyInputCodecParameters = 1768121456  # 'icdp'
        kAudioConverterPropertyOutputCodecParameters = 1868784752  # 'ocdp'
        kAudioConverterSampleRateConverterComplexity = 1936876385  # 'srca'
        kAudioConverterSampleRateConverterQuality = 1936876401  # 'srcq'
        kAudioConverterSampleRateConverterInitialPhase = 1936876400  # 'srcp'
        kAudioConverterCodecQuality = 1667527029  # 'cdqu'
        kAudioConverterPrimeMethod = 1886547309  # 'prmm'
        kAudioConverterPrimeInfo = 1886546285  # 'prim'
        kAudioConverterChannelMap = 1667788144  # 'chmp'
        kAudioConverterDecompressionMagicCookie = 1684891491  # 'dmgc'
        kAudioConverterCompressionMagicCookie = 1668114275  # 'cmgc'
        kAudioConverterEncodeBitRate = 1651663220  # 'brat'
        kAudioConverterEncodeAdjustableSampleRate = 1634366322  # 'ajsr'
        kAudioConverterInputChannelLayout = 1768123424  # 'icl '
        kAudioConverterOutputChannelLayout = 1868786720  # 'ocl '
        kAudioConverterApplicableEncodeBitRates = 1634034290  # 'aebr'
        kAudioConverterAvailableEncodeBitRates = 1986355826  # 'vebr'
        kAudioConverterApplicableEncodeSampleRates = 1634038642  # 'aesr'
        kAudioConverterAvailableEncodeSampleRates = 1986360178  # 'vesr'
        kAudioConverterAvailableEncodeChannelLayoutTags = 1634034540  # 'aecl'
        kAudioConverterCurrentOutputStreamDescription = 1633906532  # 'acod'
        kAudioConverterCurrentInputStreamDescription = 1633904996  # 'acid'
        kAudioConverterPropertySettings = 1633906803  # 'acps'
        kAudioConverterPropertyBitDepthHint = 1633903204  # 'acbd'
        kAudioConverterPropertyFormatList = 1718383476  # 'flst'

    # macOS-only properties
    ctypedef enum:
        kAudioConverterPropertyDithering = 1684632680  # 'dith'
        kAudioConverterPropertyDitherBitDepth = 1684171124  # 'dbit'

    # Quality constants
    ctypedef enum:
        kAudioConverterQuality_Max = 127
        kAudioConverterQuality_High = 96
        kAudioConverterQuality_Medium = 64
        kAudioConverterQuality_Low = 32
        kAudioConverterQuality_Min = 0

    # Sample rate converter complexity
    ctypedef enum:
        kAudioConverterSampleRateConverterComplexity_Linear = 1818846821  # 'line'
        kAudioConverterSampleRateConverterComplexity_Normal = 1852797541  # 'norm'
        kAudioConverterSampleRateConverterComplexity_Mastering = 1651471971  # 'bats'
        kAudioConverterSampleRateConverterComplexity_MinimumPhase = 1835622000  # 'minp'

    # Prime method constants
    ctypedef enum:
        kConverterPrimeMethod_Pre = 0
        kConverterPrimeMethod_Normal = 1
        kConverterPrimeMethod_None = 2

    # Dithering algorithms
    ctypedef enum:
        kDitherAlgorithm_TPDF = 1
        kDitherAlgorithm_NoiseShaping = 2

    # AudioConverterPrimeInfo structure
    ctypedef struct AudioConverterPrimeInfo:
        UInt32 leadingFrames
        UInt32 trailingFrames

    # Error codes
    ctypedef enum:
        kAudioConverterErr_FormatNotSupported = 1718447200  # 'fmt?'
        kAudioConverterErr_OperationNotSupported = 1869638207  # 'op??'
        kAudioConverterErr_PropertyNotSupported = 1886547824  # 'prop'
        kAudioConverterErr_InvalidInputSize = 1768845682  # 'insz'
        kAudioConverterErr_InvalidOutputSize = 1869771634  # 'otsz'
        kAudioConverterErr_UnspecifiedError = 2003395684  # 'what'
        kAudioConverterErr_BadPropertySizeError = 1937010802  # '!siz'
        kAudioConverterErr_RequiresPacketDescriptionsError = 1937010788  # '!pkd'
        kAudioConverterErr_InputSampleRateOutOfRange = 1768845682  # '!isr'
        kAudioConverterErr_OutputSampleRateOutOfRange = 1869771634  # '!osr'

    # iOS-only error codes
    ctypedef enum:
        kAudioConverterErr_HardwareInUse = 1752392805  # 'hwiu'
        kAudioConverterErr_NoHardwarePermission = 1886547821  # 'perm'

    # Callback function types
    ctypedef OSStatus (*AudioConverterComplexInputDataProc)(AudioConverterRef inAudioConverter,
                                                           UInt32* ioNumberDataPackets,
                                                           AudioBufferList* ioData,
                                                           AudioStreamPacketDescription** outDataPacketDescription,
                                                           void* inUserData)

    ctypedef OSStatus (*AudioConverterInputDataProc)(AudioConverterRef inAudioConverter,
                                                    UInt32* ioDataSize,
                                                    void** outData,
                                                    void* inUserData)

    # AudioConverter functions
    cdef OSStatus AudioConverterNew(const AudioStreamBasicDescription* inSourceFormat,
                                   const AudioStreamBasicDescription* inDestinationFormat,
                                   AudioConverterRef* outAudioConverter)

    cdef OSStatus AudioConverterNewSpecific(const AudioStreamBasicDescription* inSourceFormat,
                                           const AudioStreamBasicDescription* inDestinationFormat,
                                           UInt32 inNumberClassDescriptions,
                                           const AudioClassDescription* inClassDescriptions,
                                           AudioConverterRef* outAudioConverter)

    cdef OSStatus AudioConverterDispose(AudioConverterRef inAudioConverter)
    cdef OSStatus AudioConverterReset(AudioConverterRef inAudioConverter)

    cdef OSStatus AudioConverterGetPropertyInfo(AudioConverterRef inAudioConverter,
                                               AudioConverterPropertyID inPropertyID,
                                               UInt32* outSize,
                                               Boolean* outWritable)

    cdef OSStatus AudioConverterGetProperty(AudioConverterRef inAudioConverter,
                                           AudioConverterPropertyID inPropertyID,
                                           UInt32* ioPropertyDataSize,
                                           void* outPropertyData)

    cdef OSStatus AudioConverterSetProperty(AudioConverterRef inAudioConverter,
                                           AudioConverterPropertyID inPropertyID,
                                           UInt32 inPropertyDataSize,
                                           const void* inPropertyData)

    cdef OSStatus AudioConverterConvertBuffer(AudioConverterRef inAudioConverter,
                                             UInt32 inInputDataSize,
                                             const void* inInputData,
                                             UInt32* ioOutputDataSize,
                                             void* outOutputData)

    cdef OSStatus AudioConverterFillComplexBuffer(AudioConverterRef inAudioConverter,
                                                 AudioConverterComplexInputDataProc inInputDataProc,
                                                 void* inInputDataProcUserData,
                                                 UInt32* ioOutputDataPacketSize,
                                                 AudioBufferList* outOutputData,
                                                 AudioStreamPacketDescription* outPacketDescription)

    cdef OSStatus AudioConverterConvertComplexBuffer(AudioConverterRef inAudioConverter,
                                                    UInt32 inNumberPCMFrames,
                                                    const AudioBufferList* inInputData,
                                                    AudioBufferList* outOutputData)

    # DEPRECATED: Use AudioConverterFillComplexBuffer instead (macOS 10.2+)
    # This function only works with non-interleaved audio and has been superseded
    # by AudioConverterFillComplexBuffer which handles all audio configurations.
    cdef OSStatus AudioConverterFillBuffer(AudioConverterRef inAudioConverter,
                                          AudioConverterInputDataProc inInputDataProc,
                                          void* inInputDataProcUserData,
                                          UInt32* ioOutputDataSize,
                                          void* outOutputData)


cdef extern from "AudioToolbox/ExtendedAudioFile.h":

    # ExtendedAudioFile types
    ctypedef struct OpaqueExtAudioFile
    ctypedef OpaqueExtAudioFile* ExtAudioFileRef
    ctypedef UInt32 ExtAudioFilePropertyID
    ctypedef SInt32 ExtAudioFilePacketTableInfoOverride

    # Packet table info override constants
    ctypedef enum:
        kExtAudioFilePacketTableInfoOverride_UseFileValue = -1
        kExtAudioFilePacketTableInfoOverride_UseFileValueIfValid = -2

    # ExtendedAudioFile property IDs
    ctypedef enum:
        kExtAudioFileProperty_FileDataFormat = 1717988724  # 'ffmt'
        kExtAudioFileProperty_FileChannelLayout = 1717791855  # 'fclo'
        kExtAudioFileProperty_ClientDataFormat = 1667657076  # 'cfmt'
        kExtAudioFileProperty_ClientChannelLayout = 1667460207  # 'cclo'
        kExtAudioFileProperty_CodecManufacturer = 1668112750  # 'cman'
        kExtAudioFileProperty_AudioConverter = 1633906294  # 'acnv'
        kExtAudioFileProperty_AudioFile = 1634101612  # 'afil'
        kExtAudioFileProperty_FileMaxPacketSize = 1718448243  # 'fmps'
        kExtAudioFileProperty_ClientMaxPacketSize = 1668116595  # 'cmps'
        kExtAudioFileProperty_FileLengthFrames = 593916525  # '#frm'
        kExtAudioFileProperty_ConverterConfig = 1633903462  # 'accf'
        kExtAudioFileProperty_IOBufferSizeBytes = 1768907379  # 'iobs'
        kExtAudioFileProperty_IOBuffer = 1768907366  # 'iobf'
        kExtAudioFileProperty_PacketTable = 2020635753  # 'xpti'

    # Error codes
    ctypedef enum:
        kExtAudioFileError_InvalidProperty = -66561
        kExtAudioFileError_InvalidPropertySize = -66562
        kExtAudioFileError_NonPCMClientFormat = -66563
        kExtAudioFileError_InvalidChannelMap = -66564
        kExtAudioFileError_InvalidOperationOrder = -66565
        kExtAudioFileError_InvalidDataFormat = -66566
        kExtAudioFileError_MaxPacketSizeUnknown = -66567
        kExtAudioFileError_InvalidSeek = -66568
        kExtAudioFileError_AsyncWriteTooLarge = -66569
        kExtAudioFileError_AsyncWriteBufferOverflow = -66570

    # iOS-only error codes
    ctypedef enum:
        kExtAudioFileError_CodecUnavailableInputConsumed = -66559
        kExtAudioFileError_CodecUnavailableInputNotConsumed = -66560

    # ExtendedAudioFile functions
    cdef OSStatus ExtAudioFileOpenURL(CFURLRef inURL,
                                     ExtAudioFileRef* outExtAudioFile)

    cdef OSStatus ExtAudioFileWrapAudioFileID(AudioFileID inFileID,
                                             Boolean inForWriting,
                                             ExtAudioFileRef* outExtAudioFile)

    cdef OSStatus ExtAudioFileCreateWithURL(CFURLRef inURL,
                                           AudioFileTypeID inFileType,
                                           const AudioStreamBasicDescription* inStreamDesc,
                                           const AudioChannelLayout* inChannelLayout,
                                           UInt32 inFlags,
                                           ExtAudioFileRef* outExtAudioFile)

    # I/O functions
    cdef OSStatus ExtAudioFileRead(ExtAudioFileRef inExtAudioFile,
                                  UInt32* ioNumberFrames,
                                  AudioBufferList* ioData)

    cdef OSStatus ExtAudioFileWrite(ExtAudioFileRef inExtAudioFile,
                                   UInt32 inNumberFrames,
                                   const AudioBufferList* ioData)

    cdef OSStatus ExtAudioFileWriteAsync(ExtAudioFileRef inExtAudioFile,
                                        UInt32 inNumberFrames,
                                        const AudioBufferList* ioData)

    cdef OSStatus ExtAudioFileSeek(ExtAudioFileRef inExtAudioFile,
                                  SInt64 inFrameOffset)

    cdef OSStatus ExtAudioFileTell(ExtAudioFileRef inExtAudioFile,
                                  SInt64* outFrameOffset)

    # Property functions
    cdef OSStatus ExtAudioFileGetPropertyInfo(ExtAudioFileRef inExtAudioFile,
                                             ExtAudioFilePropertyID inPropertyID,
                                             UInt32* outSize,
                                             Boolean* outWritable)

    cdef OSStatus ExtAudioFileGetProperty(ExtAudioFileRef inExtAudioFile,
                                         ExtAudioFilePropertyID inPropertyID,
                                         UInt32* ioPropertyDataSize,
                                         void* outPropertyData)

    cdef OSStatus ExtAudioFileSetProperty(ExtAudioFileRef inExtAudioFile,
                                         ExtAudioFilePropertyID inPropertyID,
                                         UInt32 inPropertyDataSize,
                                         const void* inPropertyData)

    cdef OSStatus ExtAudioFileDispose(ExtAudioFileRef inExtAudioFile)


cdef extern from "AudioToolbox/AudioFormat.h":

    # AudioFormat types
    ctypedef UInt32 AudioFormatPropertyID
    ctypedef UInt32 AudioPanningMode
    ctypedef UInt32 AudioBalanceFadeType

    # Panning mode constants
    ctypedef enum:
        kPanningMode_SoundField = 3
        kPanningMode_VectorBasedPanning = 4

    # Balance fade type constants
    ctypedef enum:
        kAudioBalanceFadeType_MaxUnityGain = 0
        kAudioBalanceFadeType_EqualPower = 1

    # AudioPanningInfo structure
    ctypedef struct AudioPanningInfo:
        AudioPanningMode mPanningMode
        UInt32 mCoordinateFlags
        Float32 mCoordinates[3]
        Float32 mGainScale
        const AudioChannelLayout* mOutputChannelMap

    # AudioBalanceFade structure
    ctypedef struct AudioBalanceFade:
        Float32 mLeftRightBalance
        Float32 mBackFrontFade
        AudioBalanceFadeType mType
        const AudioChannelLayout* mChannelLayout

    # AudioFormatInfo structure
    ctypedef struct AudioFormatInfo:
        AudioStreamBasicDescription mASBD
        const void* mMagicCookie
        UInt32 mMagicCookieSize

    # ExtendedAudioFormatInfo structure
    ctypedef struct ExtendedAudioFormatInfo:
        AudioStreamBasicDescription mASBD
        const void* mMagicCookie
        UInt32 mMagicCookieSize
        AudioClassDescription mClassDescription

    # AudioFormat property IDs
    ctypedef enum:
        kAudioFormatProperty_FormatInfo = 1718449257  # 'fmti'
        kAudioFormatProperty_FormatName = 1718509933  # 'fnam'
        kAudioFormatProperty_EncodeFormatIDs = 1633906534  # 'acof'
        kAudioFormatProperty_DecodeFormatIDs = 1633904998  # 'acif'
        kAudioFormatProperty_FormatList = 1718383476  # 'flst'
        kAudioFormatProperty_ASBDFromESDS = 1702064996  # 'essd'
        kAudioFormatProperty_ChannelLayoutFromESDS = 1702060908  # 'escl'
        kAudioFormatProperty_OutputFormatList = 1868983411  # 'ofls'
        kAudioFormatProperty_FirstPlayableFormatFromList = 1718642284  # 'fpfl'
        kAudioFormatProperty_FormatIsVBR = 1719034482  # 'fvbr'
        kAudioFormatProperty_FormatIsExternallyFramed = 1717925990  # 'fexf'
        kAudioFormatProperty_FormatEmploysDependentPackets = 1717855600  # 'fdep'
        kAudioFormatProperty_FormatIsEncrypted = 1668446576  # 'cryp'
        kAudioFormatProperty_Encoders = 1635149166  # 'aven'
        kAudioFormatProperty_Decoders = 1635148901  # 'avde'
        kAudioFormatProperty_AvailableEncodeBitRates = 1634034290  # 'aebr'
        kAudioFormatProperty_AvailableEncodeSampleRates = 1634038642  # 'aesr'
        kAudioFormatProperty_AvailableEncodeChannelLayoutTags = 1634034540  # 'aecl'
        kAudioFormatProperty_AvailableEncodeNumberChannels = 1635151459  # 'avnc'
        kAudioFormatProperty_AvailableDecodeNumberChannels = 1633971811  # 'adnc'
        kAudioFormatProperty_ASBDFromMPEGPacket = 1633971568  # 'admp'
        kAudioFormatProperty_BitmapForLayoutTag = 1651340391  # 'bmtg'
        kAudioFormatProperty_MatrixMixMap = 1835884912  # 'mmap'
        kAudioFormatProperty_ChannelMap = 1667788144  # 'chmp'
        kAudioFormatProperty_NumberOfChannelsForLayout = 1852008557  # 'nchm'
        kAudioFormatProperty_AreChannelLayoutsEquivalent = 1667786097  # 'cheq'
        kAudioFormatProperty_ChannelLayoutHash = 1667786849  # 'chha'
        kAudioFormatProperty_ValidateChannelLayout = 1986093932  # 'vacl'
        kAudioFormatProperty_ChannelLayoutForTag = 1668116588  # 'cmpl'
        kAudioFormatProperty_TagForChannelLayout = 1668116596  # 'cmpt'
        kAudioFormatProperty_ChannelLayoutName = 1819242093  # 'lonm'
        kAudioFormatProperty_ChannelLayoutSimpleName = 1819504237  # 'lsnm'
        kAudioFormatProperty_ChannelLayoutForBitmap = 1668116578  # 'cmpb'
        kAudioFormatProperty_ChannelName = 1668178285  # 'cnam'
        kAudioFormatProperty_ChannelShortName = 1668509293  # 'csnm'
        kAudioFormatProperty_TagsForNumberOfChannels = 1952540515  # 'tagc'
        kAudioFormatProperty_PanningMatrix = 1885433453  # 'panm'
        kAudioFormatProperty_BalanceFade = 1650551910  # 'balf'
        kAudioFormatProperty_ID3TagSize = 1768174451  # 'id3s'
        kAudioFormatProperty_ID3TagToDictionary = 1768174436  # 'id3d'

    # iOS-only properties
    ctypedef enum:
        kAudioFormatProperty_HardwareCodecCapabilities = 1752654691  # 'hwcc'

    # iOS-only codec types
    ctypedef enum:
        kAudioDecoderComponentType = 1633969507  # 'adec'
        kAudioEncoderComponentType = 1634037347  # 'aenc'

    # iOS-only codec manufacturers
    ctypedef enum:
        kAppleSoftwareAudioCodecManufacturer = 1634758764  # 'appl'
        kAppleHardwareAudioCodecManufacturer = 1634756727  # 'aphw'

    # Error codes
    ctypedef enum:
        kAudioFormatUnspecifiedError = 2003395684  # 'what'
        kAudioFormatUnsupportedPropertyError = 1886547824  # 'prop'
        kAudioFormatBadPropertySizeError = 561211770  # '!siz'
        kAudioFormatBadSpecifierSizeError = 561213539  # '!spc'
        kAudioFormatUnsupportedDataFormatError = 1718449215  # 'fmt?'
        kAudioFormatUnknownFormatError = 560360820  # '!fmt'

    # AudioFormat functions
    cdef OSStatus AudioFormatGetPropertyInfo(AudioFormatPropertyID inPropertyID,
                                            UInt32 inSpecifierSize,
                                            const void* inSpecifier,
                                            UInt32* outPropertyDataSize)

    cdef OSStatus AudioFormatGetProperty(AudioFormatPropertyID inPropertyID,
                                        UInt32 inSpecifierSize,
                                        const void* inSpecifier,
                                        UInt32* ioPropertyDataSize,
                                        void* outPropertyData)


cdef extern from "AudioToolbox/AudioServices.h":

    # AudioServices types
    ctypedef UInt32 SystemSoundID
    ctypedef UInt32 AudioServicesPropertyID

    # AudioServices callback function type
    ctypedef void (*AudioServicesSystemSoundCompletionProc)(
        SystemSoundID ssID,
        void* clientData)

    # AudioServices error codes
    ctypedef enum:
        kAudioServicesNoError = 0
        kAudioServicesUnsupportedPropertyError = 1886547839  # 'pty?'
        kAudioServicesBadPropertySizeError = 561211770  # '!siz'
        kAudioServicesBadSpecifierSizeError = 561213539  # '!spc'
        kAudioServicesSystemSoundUnspecifiedError = -1500
        kAudioServicesSystemSoundClientTimedOutError = -1501
        kAudioServicesSystemSoundExceededMaximumDurationError = -1502

    # AudioServices system sound constants
    ctypedef enum:
        kSystemSoundID_UserPreferredAlert = 0x00001000
        kSystemSoundID_FlashScreen = 0x00000FFE
        kSystemSoundID_Vibrate = 0x00000FFF
        kUserPreferredAlert = 0x00001000  # alias for kSystemSoundID_UserPreferredAlert

    # AudioServices property IDs
    ctypedef enum:
        kAudioServicesPropertyIsUISound = 1769239657  # 'isui'
        kAudioServicesPropertyCompletePlaybackIfAppDies = 1768842857  # 'ifdi'

    # AudioServices functions
    cdef OSStatus AudioServicesCreateSystemSoundID(
        CFURLRef inFileURL,
        SystemSoundID* outSystemSoundID)

    cdef OSStatus AudioServicesDisposeSystemSoundID(SystemSoundID inSystemSoundID)

    # Note: Block-based API - actual signature uses dispatch_block_t (void (^)(void))
    # Simplified to void* for Cython compatibility since blocks cannot be directly
    # represented in Cython. Use AudioServicesAddSystemSoundCompletion for callback support.
    cdef void AudioServicesPlayAlertSoundWithCompletion(
        SystemSoundID inSystemSoundID,
        void* inCompletionBlock)

    # Note: Block-based API - actual signature uses dispatch_block_t (void (^)(void))
    # Simplified to void* for Cython compatibility since blocks cannot be directly
    # represented in Cython. Use AudioServicesAddSystemSoundCompletion for callback support.
    cdef void AudioServicesPlaySystemSoundWithCompletion(
        SystemSoundID inSystemSoundID,
        void* inCompletionBlock)

    cdef OSStatus AudioServicesGetPropertyInfo(
        AudioServicesPropertyID inPropertyID,
        UInt32 inSpecifierSize,
        const void* inSpecifier,
        UInt32* outPropertyDataSize,
        Boolean* outWritable)

    cdef OSStatus AudioServicesGetProperty(
        AudioServicesPropertyID inPropertyID,
        UInt32 inSpecifierSize,
        const void* inSpecifier,
        UInt32* ioPropertyDataSize,
        void* outPropertyData)

    cdef OSStatus AudioServicesSetProperty(
        AudioServicesPropertyID inPropertyID,
        UInt32 inSpecifierSize,
        const void* inSpecifier,
        UInt32 inPropertyDataSize,
        const void* inPropertyData)

    # DEPRECATED: Use AudioServicesPlayAlertSoundWithCompletion instead (macOS 10.11+)
    # These functions are deprecated but still widely used for simple sound playback.
    # They work synchronously and don't provide completion notification.
    cdef void AudioServicesPlayAlertSound(SystemSoundID inSystemSoundID)

    # DEPRECATED: Use AudioServicesPlaySystemSoundWithCompletion instead (macOS 10.11+)
    cdef void AudioServicesPlaySystemSound(SystemSoundID inSystemSoundID)

    cdef OSStatus AudioServicesAddSystemSoundCompletion(
        SystemSoundID inSystemSoundID,
        CFRunLoopRef inRunLoop,
        CFStringRef inRunLoopMode,
        AudioServicesSystemSoundCompletionProc inCompletionRoutine,
        void* inClientData)

    cdef void AudioServicesRemoveSystemSoundCompletion(SystemSoundID inSystemSoundID)


# MusicDevice API declarations
cdef extern from "AudioToolbox/MusicDevice.h":

    # Type definitions
    ctypedef UInt32 MusicDeviceInstrumentID
    ctypedef UInt32 MusicDeviceGroupID
    ctypedef UInt32 NoteInstanceID
    ctypedef AudioComponentInstance MusicDeviceComponent

    # Forward declaration for MIDIEventList
    ctypedef struct MIDIEventList:
        pass

    # Standard note parameters structure
    ctypedef struct MusicDeviceStdNoteParams:
        UInt32 argCount
        Float32 mPitch
        Float32 mVelocity

    # Note parameter control value
    ctypedef struct NoteParamsControlValue:
        AudioUnitParameterID mID
        AudioUnitParameterValue mValue

    # Music device note parameters (variable length)
    ctypedef struct MusicDeviceNoteParams:
        UInt32 argCount
        Float32 mPitch
        Float32 mVelocity
        NoteParamsControlValue mControls[1]

    # Constants
    ctypedef enum:
        kMusicNoteEvent_UseGroupInstrument = 0xFFFFFFFF
        kMusicNoteEvent_Unused = 0xFFFFFFFF

    # Selector constants
    ctypedef enum:
        kMusicDeviceRange = 0x0100
        kMusicDeviceMIDIEventSelect = 0x0101
        kMusicDeviceSysExSelect = 0x0102
        kMusicDevicePrepareInstrumentSelect = 0x0103
        kMusicDeviceReleaseInstrumentSelect = 0x0104
        kMusicDeviceStartNoteSelect = 0x0105
        kMusicDeviceStopNoteSelect = 0x0106
        kMusicDeviceMIDIEventListSelect = 0x0107

    # Core functions
    cdef OSStatus MusicDeviceMIDIEvent(
        MusicDeviceComponent inUnit,
        UInt32 inStatus,
        UInt32 inData1,
        UInt32 inData2,
        UInt32 inOffsetSampleFrame)

    cdef OSStatus MusicDeviceSysEx(
        MusicDeviceComponent inUnit,
        const UInt8* inData,
        UInt32 inLength)

    cdef OSStatus MusicDeviceMIDIEventList(
        MusicDeviceComponent inUnit,
        UInt32 inOffsetSampleFrame,
        const MIDIEventList* evtList)

    cdef OSStatus MusicDeviceStartNote(
        MusicDeviceComponent inUnit,
        MusicDeviceInstrumentID inInstrument,
        MusicDeviceGroupID inGroupID,
        NoteInstanceID* outNoteInstanceID,
        UInt32 inOffsetSampleFrame,
        const MusicDeviceNoteParams* inParams)

    cdef OSStatus MusicDeviceStopNote(
        MusicDeviceComponent inUnit,
        MusicDeviceGroupID inGroupID,
        NoteInstanceID inNoteInstanceID,
        UInt32 inOffsetSampleFrame)

    # Function pointer types for fast dispatch
    ctypedef OSStatus (*MusicDeviceMIDIEventProc)(
        void* self,
        UInt32 inStatus,
        UInt32 inData1,
        UInt32 inData2,
        UInt32 inOffsetSampleFrame)

    ctypedef OSStatus (*MusicDeviceSysExProc)(
        void* self,
        const UInt8* inData,
        UInt32 inLength)

    ctypedef OSStatus (*MusicDeviceStartNoteProc)(
        void* self,
        MusicDeviceInstrumentID inInstrument,
        MusicDeviceGroupID inGroupID,
        NoteInstanceID* outNoteInstanceID,
        UInt32 inOffsetSampleFrame,
        const MusicDeviceNoteParams* inParams)

    ctypedef OSStatus (*MusicDeviceStopNoteProc)(
        void* self,
        MusicDeviceGroupID inGroupID,
        NoteInstanceID inNoteInstanceID,
        UInt32 inOffsetSampleFrame)


# MusicPlayer API declarations
cdef extern from "AudioToolbox/MusicPlayer.h":

    # Opaque types
    ctypedef struct OpaqueMusicPlayer:
        pass
    ctypedef OpaqueMusicPlayer* MusicPlayer

    ctypedef struct OpaqueMusicSequence:
        pass
    ctypedef OpaqueMusicSequence* MusicSequence

    ctypedef struct OpaqueMusicTrack:
        pass
    ctypedef OpaqueMusicTrack* MusicTrack

    ctypedef struct OpaqueMusicEventIterator:
        pass
    ctypedef OpaqueMusicEventIterator* MusicEventIterator

    # Basic types
    ctypedef Float64 MusicTimeStamp
    ctypedef UInt32 MusicEventType

    # Forward declarations for missing types
    ctypedef void* AUGraph
    ctypedef UInt32 AUNode

    # Event types constants
    ctypedef enum:
        kMusicEventType_NULL = 0
        kMusicEventType_ExtendedNote = 1
        kMusicEventType_ExtendedTempo = 3
        kMusicEventType_User = 4
        kMusicEventType_Meta = 5
        kMusicEventType_MIDINoteMessage = 6
        kMusicEventType_MIDIChannelMessage = 7
        kMusicEventType_MIDIRawData = 8
        kMusicEventType_Parameter = 9
        kMusicEventType_AUPreset = 10

    # Sequence types
    ctypedef enum:
        kMusicSequenceType_Beats = 1651077476  # 'beat'
        kMusicSequenceType_Seconds = 1936941667  # 'secs'
        kMusicSequenceType_Samples = 1935832176  # 'samp'

    # File types
    ctypedef enum:
        kMusicSequenceFile_AnyType = 0
        kMusicSequenceFile_MIDIType = 1835623529  # 'midi'
        kMusicSequenceFile_iMelodyType = 1768776044  # 'imel'

    # Load and file flags
    ctypedef enum:
        kMusicSequenceLoadSMF_PreserveTracks = 0
        kMusicSequenceLoadSMF_ChannelsToTracks = 1
        kMusicSequenceFileFlags_Default = 0
        kMusicSequenceFileFlags_EraseFile = 1

    # Track property constants
    ctypedef enum:
        kSequenceTrackProperty_LoopInfo = 0
        kSequenceTrackProperty_OffsetTime = 1
        kSequenceTrackProperty_MuteStatus = 2
        kSequenceTrackProperty_SoloStatus = 3
        kSequenceTrackProperty_AutomatedParameters = 4
        kSequenceTrackProperty_TrackLength = 5
        kSequenceTrackProperty_TimeResolution = 6

    # Error constants
    ctypedef enum:
        kAudioToolboxErr_InvalidSequenceType = -10846
        kAudioToolboxErr_TrackIndexError = -10859
        kAudioToolboxErr_TrackNotFound = -10858
        kAudioToolboxErr_EndOfTrack = -10857
        kAudioToolboxErr_StartOfTrack = -10856
        kAudioToolboxErr_IllegalTrackDestination = -10855
        kAudioToolboxErr_NoSequence = -10854
        kAudioToolboxErr_InvalidEventType = -10853
        kAudioToolboxErr_InvalidPlayerState = -10852
        kAudioToolboxErr_CannotDoInCurrentContext = -10863
        kAudioToolboxError_NoTrackDestination = -66720

    # Event structures
    ctypedef struct MIDINoteMessage:
        UInt8 channel
        UInt8 note
        UInt8 velocity
        UInt8 releaseVelocity
        Float32 duration

    ctypedef struct MIDIChannelMessage:
        UInt8 status
        UInt8 data1
        UInt8 data2
        UInt8 reserved

    ctypedef struct MIDIRawData:
        UInt32 length
        UInt8 data[1]

    ctypedef struct MIDIMetaEvent:
        UInt8 metaEventType
        UInt8 unused1
        UInt8 unused2
        UInt8 unused3
        UInt32 dataLength
        UInt8 data[1]

    ctypedef struct MusicEventUserData:
        UInt32 length
        UInt8 data[1]

    ctypedef struct ExtendedNoteOnEvent:
        MusicDeviceInstrumentID instrumentID
        MusicDeviceGroupID groupID
        Float32 duration
        MusicDeviceNoteParams extendedParams

    ctypedef struct ParameterEvent:
        AudioUnitParameterID parameterID
        AudioUnitScope scope
        AudioUnitElement element
        AudioUnitParameterValue value

    ctypedef struct ExtendedTempoEvent:
        Float64 bpm

    ctypedef struct AUPresetEvent:
        AudioUnitScope scope
        AudioUnitElement element
        void* preset  # CFPropertyListRef

    ctypedef struct MusicTrackLoopInfo:
        MusicTimeStamp loopDuration
        SInt32 numberOfLoops

    ctypedef struct CABarBeatTime:
        SInt32 bar
        UInt16 beat
        UInt16 subbeat
        UInt16 subbeatDivisor
        UInt16 reserved

    # Callback types
    ctypedef void (*MusicSequenceUserCallback)(
        void* inClientData,
        MusicSequence inSequence,
        MusicTrack inTrack,
        MusicTimeStamp inEventTime,
        const MusicEventUserData* inEventData,
        MusicTimeStamp inStartSliceBeat,
        MusicTimeStamp inEndSliceBeat)

    # MusicPlayer functions
    cdef OSStatus NewMusicPlayer(MusicPlayer* outPlayer)
    cdef OSStatus DisposeMusicPlayer(MusicPlayer inPlayer)
    cdef OSStatus MusicPlayerSetSequence(MusicPlayer inPlayer, MusicSequence inSequence)
    cdef OSStatus MusicPlayerGetSequence(MusicPlayer inPlayer, MusicSequence* outSequence)
    cdef OSStatus MusicPlayerSetTime(MusicPlayer inPlayer, MusicTimeStamp inTime)
    cdef OSStatus MusicPlayerGetTime(MusicPlayer inPlayer, MusicTimeStamp* outTime)
    cdef OSStatus MusicPlayerGetHostTimeForBeats(MusicPlayer inPlayer, MusicTimeStamp inBeats, UInt64* outHostTime)
    cdef OSStatus MusicPlayerGetBeatsForHostTime(MusicPlayer inPlayer, UInt64 inHostTime, MusicTimeStamp* outBeats)
    cdef OSStatus MusicPlayerPreroll(MusicPlayer inPlayer)
    cdef OSStatus MusicPlayerStart(MusicPlayer inPlayer)
    cdef OSStatus MusicPlayerStop(MusicPlayer inPlayer)
    cdef OSStatus MusicPlayerIsPlaying(MusicPlayer inPlayer, Boolean* outIsPlaying)
    cdef OSStatus MusicPlayerSetPlayRateScalar(MusicPlayer inPlayer, Float64 inScaleRate)
    cdef OSStatus MusicPlayerGetPlayRateScalar(MusicPlayer inPlayer, Float64* outScaleRate)

    # MusicSequence functions
    cdef OSStatus NewMusicSequence(MusicSequence* outSequence)
    cdef OSStatus DisposeMusicSequence(MusicSequence inSequence)
    cdef OSStatus MusicSequenceNewTrack(MusicSequence inSequence, MusicTrack* outTrack)
    cdef OSStatus MusicSequenceDisposeTrack(MusicSequence inSequence, MusicTrack inTrack)
    cdef OSStatus MusicSequenceGetTrackCount(MusicSequence inSequence, UInt32* outNumberOfTracks)
    cdef OSStatus MusicSequenceGetIndTrack(MusicSequence inSequence, UInt32 inTrackIndex, MusicTrack* outTrack)
    cdef OSStatus MusicSequenceGetTrackIndex(MusicSequence inSequence, MusicTrack inTrack, UInt32* outTrackIndex)
    cdef OSStatus MusicSequenceGetTempoTrack(MusicSequence inSequence, MusicTrack* outTrack)
    cdef OSStatus MusicSequenceSetAUGraph(MusicSequence inSequence, AUGraph inGraph)
    cdef OSStatus MusicSequenceGetAUGraph(MusicSequence inSequence, AUGraph* outGraph)
    cdef OSStatus MusicSequenceSetSequenceType(MusicSequence inSequence, UInt32 inType)
    cdef OSStatus MusicSequenceGetSequenceType(MusicSequence inSequence, UInt32* outType)
    cdef OSStatus MusicSequenceFileLoad(MusicSequence inSequence, CFURLRef inFileRef, UInt32 inFileTypeHint, UInt32 inFlags)
    # cdef OSStatus MusicSequenceFileLoadData(MusicSequence inSequence, CFDataRef inData, UInt32 inFileTypeHint, UInt32 inFlags)
    # cdef OSStatus MusicSequenceFileCreate(MusicSequence inSequence, CFURLRef inFileRef, UInt32 inFileType, UInt32 inFlags, SInt16 inResolution)
    # cdef OSStatus MusicSequenceFileCreateData(MusicSequence inSequence, UInt32 inFileType, UInt32 inFlags, SInt16 inResolution, CFDataRef* outData)
    cdef OSStatus MusicSequenceReverse(MusicSequence inSequence)
    cdef OSStatus MusicSequenceGetSecondsForBeats(MusicSequence inSequence, MusicTimeStamp inBeats, Float64* outSeconds)
    cdef OSStatus MusicSequenceGetBeatsForSeconds(MusicSequence inSequence, Float64 inSeconds, MusicTimeStamp* outBeats)
    cdef OSStatus MusicSequenceSetUserCallback(MusicSequence inSequence, MusicSequenceUserCallback inCallback, void* inClientData)
    cdef OSStatus MusicSequenceBeatsToBarBeatTime(MusicSequence inSequence, MusicTimeStamp inBeats, UInt32 inSubbeatDivisor, CABarBeatTime* outBarBeatTime)
    cdef OSStatus MusicSequenceBarBeatTimeToBeats(MusicSequence inSequence, const CABarBeatTime* inBarBeatTime, MusicTimeStamp* outBeats)
    # cdef CFDictionaryRef MusicSequenceGetInfoDictionary(MusicSequence inSequence)

    # MusicTrack functions
    cdef OSStatus MusicTrackGetSequence(MusicTrack inTrack, MusicSequence* outSequence)
    cdef OSStatus MusicTrackSetDestNode(MusicTrack inTrack, AUNode inNode)
    cdef OSStatus MusicTrackGetDestNode(MusicTrack inTrack, AUNode* outNode)
    cdef OSStatus MusicTrackSetProperty(MusicTrack inTrack, UInt32 inPropertyID, void* inData, UInt32 inLength)
    cdef OSStatus MusicTrackGetProperty(MusicTrack inTrack, UInt32 inPropertyID, void* outData, UInt32* ioLength)
    cdef OSStatus MusicTrackMoveEvents(MusicTrack inTrack, MusicTimeStamp inStartTime, MusicTimeStamp inEndTime, MusicTimeStamp inMoveTime)
    cdef OSStatus MusicTrackClear(MusicTrack inTrack, MusicTimeStamp inStartTime, MusicTimeStamp inEndTime)
    cdef OSStatus MusicTrackCut(MusicTrack inTrack, MusicTimeStamp inStartTime, MusicTimeStamp inEndTime)
    cdef OSStatus MusicTrackCopyInsert(MusicTrack inSourceTrack, MusicTimeStamp inSourceStartTime, MusicTimeStamp inSourceEndTime, MusicTrack inDestTrack, MusicTimeStamp inDestInsertTime)
    cdef OSStatus MusicTrackMerge(MusicTrack inSourceTrack, MusicTimeStamp inSourceStartTime, MusicTimeStamp inSourceEndTime, MusicTrack inDestTrack, MusicTimeStamp inDestInsertTime)

    # MusicTrack event creation functions
    cdef OSStatus MusicTrackNewMIDINoteEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const MIDINoteMessage* inMessage)
    cdef OSStatus MusicTrackNewMIDIChannelEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const MIDIChannelMessage* inMessage)
    cdef OSStatus MusicTrackNewMIDIRawDataEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const MIDIRawData* inRawData)
    cdef OSStatus MusicTrackNewExtendedNoteEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const ExtendedNoteOnEvent* inInfo)
    cdef OSStatus MusicTrackNewParameterEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const ParameterEvent* inInfo)
    cdef OSStatus MusicTrackNewExtendedTempoEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, Float64 inBPM)
    cdef OSStatus MusicTrackNewMetaEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const MIDIMetaEvent* inMetaEvent)
    cdef OSStatus MusicTrackNewUserEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const MusicEventUserData* inUserData)
    cdef OSStatus MusicTrackNewAUPresetEvent(MusicTrack inTrack, MusicTimeStamp inTimeStamp, const AUPresetEvent* inPresetEvent)

    # MusicEventIterator functions
    cdef OSStatus NewMusicEventIterator(MusicTrack inTrack, MusicEventIterator* outIterator)
    cdef OSStatus DisposeMusicEventIterator(MusicEventIterator inIterator)
    cdef OSStatus MusicEventIteratorSeek(MusicEventIterator inIterator, MusicTimeStamp inTimeStamp)
    cdef OSStatus MusicEventIteratorNextEvent(MusicEventIterator inIterator)
    cdef OSStatus MusicEventIteratorPreviousEvent(MusicEventIterator inIterator)
    cdef OSStatus MusicEventIteratorGetEventInfo(MusicEventIterator inIterator, MusicTimeStamp* outTimeStamp, MusicEventType* outEventType, const void** outEventData, UInt32* outEventDataSize)
    cdef OSStatus MusicEventIteratorSetEventInfo(MusicEventIterator inIterator, MusicEventType inEventType, const void* inEventData)
    cdef OSStatus MusicEventIteratorSetEventTime(MusicEventIterator inIterator, MusicTimeStamp inTimeStamp)
    cdef OSStatus MusicEventIteratorDeleteEvent(MusicEventIterator inIterator)
    cdef OSStatus MusicEventIteratorHasPreviousEvent(MusicEventIterator inIterator, Boolean* outHasPrevEvent)
    cdef OSStatus MusicEventIteratorHasNextEvent(MusicEventIterator inIterator, Boolean* outHasNextEvent)
    cdef OSStatus MusicEventIteratorHasCurrentEvent(MusicEventIterator inIterator, Boolean* outHasCurEvent)


# -----------------------------------------------------------------------------
# AUGraph API
# -----------------------------------------------------------------------------

cdef extern from "AudioToolbox/AUGraph.h":

    # Type definitions
    ctypedef struct OpaqueAUGraph
    ctypedef OpaqueAUGraph* AUGraph
    ctypedef SInt32 AUNode

    # Error codes
    ctypedef enum:
        kAUGraphErr_NodeNotFound = -10860
        kAUGraphErr_InvalidConnection = -10861
        kAUGraphErr_OutputNodeErr = -10862
        kAUGraphErr_CannotDoInCurrentContext = -10863
        kAUGraphErr_InvalidAudioUnit = -10864

    # Node interaction types
    ctypedef enum:
        kAUNodeInteraction_Connection = 1
        kAUNodeInteraction_InputCallback = 2

    # Structures
    ctypedef struct AudioUnitNodeConnection:
        AUNode sourceNode
        UInt32 sourceOutputNumber
        AUNode destNode
        UInt32 destInputNumber

    ctypedef AudioUnitNodeConnection AUNodeConnection

    ctypedef struct AUNodeRenderCallback:
        AUNode destNode
        UInt32 destInputNumber
        AURenderCallbackStruct cback

    ctypedef struct AUNodeInteraction:
        UInt32 nodeInteractionType
        # Union not fully declared - we'll handle at Python level

    # Graph lifecycle
    cdef OSStatus NewAUGraph(AUGraph* outGraph)
    cdef OSStatus DisposeAUGraph(AUGraph inGraph)
    cdef OSStatus AUGraphOpen(AUGraph inGraph)
    cdef OSStatus AUGraphClose(AUGraph inGraph)
    cdef OSStatus AUGraphInitialize(AUGraph inGraph)
    cdef OSStatus AUGraphUninitialize(AUGraph inGraph)

    # Graph control
    cdef OSStatus AUGraphStart(AUGraph inGraph)
    cdef OSStatus AUGraphStop(AUGraph inGraph)
    cdef OSStatus AUGraphIsOpen(AUGraph inGraph, Boolean* outIsOpen)
    cdef OSStatus AUGraphIsInitialized(AUGraph inGraph, Boolean* outIsInitialized)
    cdef OSStatus AUGraphIsRunning(AUGraph inGraph, Boolean* outIsRunning)

    # Node management
    cdef OSStatus AUGraphAddNode(AUGraph inGraph, const AudioComponentDescription* inDescription, AUNode* outNode)
    cdef OSStatus AUGraphRemoveNode(AUGraph inGraph, AUNode inNode)
    cdef OSStatus AUGraphGetNodeCount(AUGraph inGraph, UInt32* outNumberOfNodes)
    cdef OSStatus AUGraphGetIndNode(AUGraph inGraph, UInt32 inIndex, AUNode* outNode)
    cdef OSStatus AUGraphNodeInfo(AUGraph inGraph, AUNode inNode, AudioComponentDescription* outDescription, AudioUnit* outAudioUnit)

    # Connections
    cdef OSStatus AUGraphConnectNodeInput(AUGraph inGraph, AUNode inSourceNode, UInt32 inSourceOutputNumber, AUNode inDestNode, UInt32 inDestInputNumber)
    cdef OSStatus AUGraphDisconnectNodeInput(AUGraph inGraph, AUNode inDestNode, UInt32 inDestInputNumber)
    cdef OSStatus AUGraphClearConnections(AUGraph inGraph)

    # Graph updates
    cdef OSStatus AUGraphUpdate(AUGraph inGraph, Boolean* outIsUpdated)

    # Utilities
    cdef OSStatus AUGraphGetCPULoad(AUGraph inGraph, Float32* outAverageCPULoad)
    cdef OSStatus AUGraphGetMaxCPULoad(AUGraph inGraph, Float32* outMaxLoad)


# ============================================================================
# CoreAudioClock - Audio/MIDI Synchronization and Timing Services
# ============================================================================

cdef extern from "AudioToolbox/CoreAudioClock.h":
    # Opaque clock reference
    ctypedef struct OpaqueCAClock:
        pass
    ctypedef OpaqueCAClock* CAClockRef

    # Time and tempo types
    ctypedef Float64 CAClockBeats       # MIDI quarter notes
    ctypedef Float64 CAClockTempo       # beats per minute
    ctypedef Float64 CAClockSamples     # audio samples
    ctypedef Float64 CAClockSeconds     # seconds

    # Property IDs
    ctypedef enum CAClockPropertyID:
        kCAClockProperty_InternalTimebase       = 0x696E7462  # 'intb'
        kCAClockProperty_TimebaseSource         = 0x69746273  # 'itbs'
        kCAClockProperty_SyncMode               = 0x73796E6D  # 'synm'
        kCAClockProperty_SyncSource             = 0x73796E73  # 'syns'
        kCAClockProperty_SMPTEFormat            = 0x736D7066  # 'smpf'
        kCAClockProperty_SMPTEOffset            = 0x736D706F  # 'smpo'
        kCAClockProperty_MIDIClockDestinations  = 0x6D626364  # 'mbcd'
        kCAClockProperty_MTCDestinations        = 0x6D746364  # 'mtcd'
        kCAClockProperty_MTCFreewheelTime       = 0x6D746677  # 'mtfw'
        kCAClockProperty_TempoMap               = 0x746D706F  # 'tmpo'
        kCAClockProperty_MeterTrack             = 0x6D657472  # 'metr'
        kCAClockProperty_Name                   = 0x6E616D65  # 'name'
        kCAClockProperty_SendMIDISPP            = 0x6D737070  # 'mspp'

    # Timebase types
    ctypedef enum CAClockTimebase:
        kCAClockTimebase_HostTime           = 0x686F7374  # 'host'
        kCAClockTimebase_AudioDevice        = 0x61756469  # 'audi'
        kCAClockTimebase_AudioOutputUnit    = 0x61756F75  # 'auou'

    # Sync modes
    ctypedef enum CAClockSyncMode:
        kCAClockSyncMode_Internal               = 0x696E7472  # 'intr'
        kCAClockSyncMode_MIDIClockTransport     = 0x6D636C6B  # 'mclk'
        kCAClockSyncMode_MTCTransport           = 0x6D6D7463  # 'mmtc'

    # SMPTE format (already defined in coreaudio.pxd as SMPTETimeType)
    ctypedef SMPTETimeType CAClockSMPTEFormat

    # Clock messages
    ctypedef enum CAClockMessage:
        kCAClockMessage_StartTimeSet        = 0x7374696D  # 'stim'
        kCAClockMessage_Started             = 0x73747274  # 'strt'
        kCAClockMessage_Stopped             = 0x73746F70  # 'stop'
        kCAClockMessage_Armed               = 0x61726D64  # 'armd'
        kCAClockMessage_Disarmed            = 0x6461726D  # 'darm'
        kCAClockMessage_PropertyChanged     = 0x70636867  # 'pchg'
        kCAClockMessage_WrongSMPTEFormat    = 0x3F736D70  # '?smp'

    # Time formats
    ctypedef enum CAClockTimeFormat:
        kCAClockTimeFormat_HostTime         = 0x686F7374  # 'host'
        kCAClockTimeFormat_Samples          = 0x73616D70  # 'samp'
        kCAClockTimeFormat_Beats            = 0x62656174  # 'beat'
        kCAClockTimeFormat_Seconds          = 0x73656373  # 'secs'
        kCAClockTimeFormat_SMPTESeconds     = 0x736D7073  # 'smps'
        kCAClockTimeFormat_SMPTETime        = 0x736D7074  # 'smpt'
        kCAClockTimeFormat_AbsoluteSeconds  = 0x61736563  # 'asec'

    # Clock time structure
    ctypedef union CAClockTimeUnion:
        UInt64          hostTime
        CAClockSamples  samples
        CAClockBeats    beats
        CAClockSeconds  seconds
        SMPTETime       smpte

    ctypedef struct CAClockTime:
        CAClockTimeFormat   format
        UInt32              reserved
        CAClockTimeUnion    time

    # Tempo map entry
    ctypedef struct CATempoMapEntry:
        CAClockBeats    beats
        CAClockTempo    tempoBPM

    # Meter track entry
    ctypedef struct CAMeterTrackEntry:
        CAClockBeats    beats
        UInt16          meterNumer
        UInt16          meterDenom

    # Listener callback
    ctypedef void (*CAClockListenerProc)(void* userData, CAClockMessage message, const void* param)

    # Clock lifecycle functions
    cdef OSStatus CAClockNew(UInt32 inReservedFlags, CAClockRef* outCAClock)
    cdef OSStatus CAClockDispose(CAClockRef inCAClock)

    # Property functions
    cdef OSStatus CAClockGetPropertyInfo(CAClockRef inCAClock, CAClockPropertyID inPropertyID, UInt32* outSize, Boolean* outWritable)
    cdef OSStatus CAClockGetProperty(CAClockRef inCAClock, CAClockPropertyID inPropertyID, UInt32* ioSize, void* outData)
    cdef OSStatus CAClockSetProperty(CAClockRef inCAClock, CAClockPropertyID inPropertyID, UInt32 inSize, const void* inData)

    # Listener functions
    cdef OSStatus CAClockAddListener(CAClockRef inCAClock, CAClockListenerProc inListenerProc, void* inUserData)
    cdef OSStatus CAClockRemoveListener(CAClockRef inCAClock, CAClockListenerProc inListenerProc, void* inUserData)

    # Time control functions
    cdef OSStatus CAClockSetCurrentTime(CAClockRef inCAClock, const CAClockTime* inTime)
    cdef OSStatus CAClockGetCurrentTime(CAClockRef inCAClock, CAClockTimeFormat inFormat, CAClockTime* outTime)
    cdef OSStatus CAClockGetStartTime(CAClockRef inCAClock, CAClockTimeFormat inFormat, CAClockTime* outTime)

    # Playback control
    cdef OSStatus CAClockSetPlayRate(CAClockRef inCAClock, Float64 inRate)
    cdef OSStatus CAClockGetPlayRate(CAClockRef inCAClock, Float64* outRate)
    cdef OSStatus CAClockArm(CAClockRef inCAClock)
    cdef OSStatus CAClockDisarm(CAClockRef inCAClock)
    cdef OSStatus CAClockStart(CAClockRef inCAClock)
    cdef OSStatus CAClockStop(CAClockRef inCAClock)

    # Note: Some functions like CAClockSetCurrentTempo, CAClockSecondsToBeats may not be
    # available in all macOS versions or may have different signatures

    # Error codes
    cdef enum:
        kCAClock_UnknownPropertyError           = -66816
        kCAClock_InvalidPropertySizeError       = -66815
        kCAClock_InvalidTimeFormatError         = -66814
        kCAClock_InvalidSyncModeError           = -66813
        kCAClock_InvalidSyncSourceError         = -66812
        kCAClock_InvalidTimebaseError           = -66811
        kCAClock_InvalidTimebaseSourceError     = -66810
        kCAClock_InvalidSMPTEFormatError        = -66809
        kCAClock_InvalidSMPTEOffsetError        = -66808
        kCAClock_InvalidUnitError               = -66807
        kCAClock_InvalidPlayRateError           = -66806
        kCAClock_CannotSetTimeError             = -66805



#!/usr/bin/env python3
"""CoreMusic audio package.

This package contains audio-related modules:
- async_io: Asynchronous I/O classes for audio operations
- streaming: Real-time audio streaming and processing
- analysis: Audio analysis and feature extraction
- slicing: Audio slicing and recombination
- visualization: Audio visualization and plotting
- utilities: High-level audio processing utilities
- mmap_file: Memory-mapped file access for performance
- buffer_pool: Buffer pooling for efficient memory management
"""

# Import analysis classes
from .analysis import AudioAnalyzer, BeatInfo, LivePitchDetector, PitchInfo
# Import async I/O classes (re-exported)
from .async_io import AsyncAudioFile as AsyncAudioFile
from .async_io import AsyncAudioQueue as AsyncAudioQueue
from .async_io import create_output_queue_async as create_output_queue_async
from .async_io import open_audio_file_async as open_audio_file_async
from .buffer_pool import (BufferPool, BufferPoolStats, PooledBuffer,
                          get_global_pool, reset_global_pool)
# Import performance modules
from .mmap_file import MMapAudioFile
# Import slicing classes
from .slicing import (AudioSlicer, RecombineMethod, Slice, SliceCollection,
                      SliceMethod, SliceRecombinator)
# Import streaming classes
from .streaming import (AudioInputStream, AudioOutputStream, AudioProcessor,
                        StreamGraph, StreamNode, create_loopback)
# Import utilities
from .utilities import (AudioEffectsChain, AudioFormatPresets, batch_convert,
                        batch_process_files, batch_process_parallel,
                        convert_audio_file, create_simple_effect_chain,
                        find_audio_unit_by_name, get_audiounit_names,
                        list_available_audio_units,
                        parse_audio_stream_basic_description, trim_audio)
# Import visualization classes
from .visualization import (MATPLOTLIB_AVAILABLE, NUMPY_AVAILABLE,
                            FrequencySpectrumPlotter, SpectrogramPlotter,
                            WaveformPlotter)

# Import Cython-optimized operations from capi (re-exported when available)
# These are defined in capi.pyx and may not be visible to type checkers
try:
    from ..capi import apply_fade_in_float32 as apply_fade_in_float32  # type: ignore[attr-defined]
    from ..capi import apply_fade_out_float32 as apply_fade_out_float32  # type: ignore[attr-defined]
    from ..capi import apply_gain as apply_gain  # type: ignore[attr-defined]
    from ..capi import apply_gain_float32 as apply_gain_float32  # type: ignore[attr-defined]
    from ..capi import calculate_peak as calculate_peak  # type: ignore[attr-defined]
    from ..capi import calculate_peak_float32 as calculate_peak_float32  # type: ignore[attr-defined]
    from ..capi import calculate_rms as calculate_rms  # type: ignore[attr-defined]
    from ..capi import calculate_rms_float32 as calculate_rms_float32  # type: ignore[attr-defined]
    from ..capi import convert_float32_to_int16 as convert_float32_to_int16  # type: ignore[attr-defined]
    from ..capi import convert_int16_to_float32 as convert_int16_to_float32  # type: ignore[attr-defined]
    from ..capi import mix_audio_float32 as mix_audio_float32  # type: ignore[attr-defined]
    from ..capi import mono_to_stereo_float32 as mono_to_stereo_float32  # type: ignore[attr-defined]
    from ..capi import normalize_audio as normalize_audio  # type: ignore[attr-defined]
    from ..capi import normalize_audio_float32 as normalize_audio_float32  # type: ignore[attr-defined]
    from ..capi import stereo_to_mono_float32 as stereo_to_mono_float32  # type: ignore[attr-defined]
    CYTHON_OPS_AVAILABLE = True
except ImportError:
    CYTHON_OPS_AVAILABLE = False

__all__ = [
    # Async I/O
    "AsyncAudioFile",
    "AsyncAudioQueue",
    # Streaming
    "AudioInputStream",
    "AudioOutputStream",
    "AudioProcessor",
    "StreamGraph",
    "StreamNode",
    "create_loopback",
    # Analysis
    "AudioAnalyzer",
    "LivePitchDetector",
    "BeatInfo",
    "PitchInfo",
    # Slicing
    "Slice",
    "AudioSlicer",
    "SliceCollection",
    "SliceRecombinator",
    "SliceMethod",
    "RecombineMethod",
    # Visualization
    "WaveformPlotter",
    "SpectrogramPlotter",
    "FrequencySpectrumPlotter",
    "MATPLOTLIB_AVAILABLE",
    "NUMPY_AVAILABLE",
    # Utilities
    "AudioEffectsChain",
    "AudioFormatPresets",
    "batch_convert",
    "batch_process_files",
    "batch_process_parallel",
    "convert_audio_file",
    "create_simple_effect_chain",
    "find_audio_unit_by_name",
    "get_audiounit_names",
    "list_available_audio_units",
    "parse_audio_stream_basic_description",
    "trim_audio",
    # Performance
    "MMapAudioFile",
    "BufferPool",
    "PooledBuffer",
    "BufferPoolStats",
    "get_global_pool",
    "reset_global_pool",
    # Cython ops (if available)
    "CYTHON_OPS_AVAILABLE",
]

# Add Cython ops to __all__ if available
if CYTHON_OPS_AVAILABLE:
    __all__.extend([
        "normalize_audio",
        "apply_gain",
        "calculate_rms",
        "calculate_peak",
        "normalize_audio_float32",
        "apply_gain_float32",
        "mix_audio_float32",
        "convert_float32_to_int16",
        "convert_int16_to_float32",
        "stereo_to_mono_float32",
        "mono_to_stereo_float32",
        "apply_fade_in_float32",
        "apply_fade_out_float32",
        "calculate_rms_float32",
        "calculate_peak_float32",
    ])



# corefoundation.pxd
# CoreFoundation framework declarations for coremusic

cdef extern from *:
    """
    #define FOURCC_ARGS(x)  (char)((x & 0xff000000) >> 24), \
        (char)((x & 0xff0000) >> 16),                   \
        (char)((x & 0xff00) >> 8), (char)((x) & 0xff)
    """
    ctypedef unsigned long uint64_t
    ctypedef long int64_t
    cdef char[] FOURCC_ARGS(SInt32)

# -----------------------------------------------------------------------------

cdef extern from "CoreFoundation/CFBase.h":
    ctypedef float          Float32
    ctypedef double         Float64
    ctypedef unsigned char  Boolean
    ctypedef unsigned char  UInt8
    ctypedef signed char    SInt8
    ctypedef unsigned short UInt16
    ctypedef signed short   SInt16
    ctypedef unsigned int   UInt32
    ctypedef signed int     SInt32
    ctypedef uint64_t       UInt64
    ctypedef int64_t        SInt64
    ctypedef SInt32         OSStatus

    ctypedef UInt32          FourCharCode
    ctypedef FourCharCode    OSType

cdef extern from "CoreFoundation/CoreFoundation.h":
    ctypedef struct __CFURL
    ctypedef __CFURL* CFURLRef
    ctypedef struct __CFRunLoop
    ctypedef __CFRunLoop* CFRunLoopRef
    ctypedef struct __CFString
    ctypedef __CFString* CFStringRef
    ctypedef struct __CFData
    ctypedef __CFData* CFDataRef
    ctypedef UInt8 AudioFilePermissions
    ctypedef long CFIndex
    ctypedef void* CFAllocatorRef
    ctypedef void* CFTypeRef

    cdef CFURLRef CFURLCreateFromFileSystemRepresentation(CFAllocatorRef allocator, const UInt8* buffer, CFIndex bufLen, Boolean isDirectory)

    # CFURL path style constants
    ctypedef enum CFURLPathStyle:
        kCFURLPOSIXPathStyle = 0
        kCFURLHFSPathStyle = 1
        kCFURLWindowsPathStyle = 2

    cdef CFURLRef CFURLCreateWithFileSystemPath(CFAllocatorRef allocator, CFStringRef filePath, CFURLPathStyle pathStyle, Boolean isDirectory)

    cdef void CFRelease(CFTypeRef cf)
    cdef CFAllocatorRef kCFAllocatorDefault

    # CFRunLoop functions for audio queue callbacks
    cdef CFRunLoopRef CFRunLoopGetCurrent()
    cdef SInt32 CFRunLoopRunInMode(CFStringRef mode, double seconds, Boolean returnAfterSourceHandled)
    cdef CFStringRef kCFRunLoopDefaultMode
    cdef CFStringRef kCFRunLoopCommonModes

    # CFRunLoopRunInMode return values
    ctypedef enum:
        kCFRunLoopRunFinished = 1
        kCFRunLoopRunStopped = 2
        kCFRunLoopRunTimedOut = 3
        kCFRunLoopRunHandledSource = 4

    # Type checking functions
    ctypedef unsigned long CFTypeID
    cdef CFTypeID CFGetTypeID(CFTypeRef cf)
    cdef CFTypeID CFStringGetTypeID()
    cdef CFTypeID CFNumberGetTypeID()

    # String creation and encoding
    ctypedef UInt32 CFStringEncoding
    cdef CFStringEncoding kCFStringEncodingUTF8
    cdef CFStringRef CFStringCreateWithCString(CFAllocatorRef allocator, const char* cStr, CFStringEncoding encoding)
    cdef char* CFStringGetCStringPtr(CFStringRef theString, CFStringEncoding encoding)
    cdef CFIndex CFStringGetLength(CFStringRef theString)
    cdef CFIndex CFStringGetMaximumSizeForEncoding(CFIndex length, CFStringEncoding encoding)
    cdef Boolean CFStringGetCString(CFStringRef theString, char* buffer, CFIndex bufferSize, CFStringEncoding encoding)

    # CFData functions
    cdef CFDataRef CFDataCreate(CFAllocatorRef allocator, const UInt8* bytes, CFIndex length)
    cdef CFIndex CFDataGetLength(CFDataRef theData)
    cdef UInt8* CFDataGetBytePtr(CFDataRef theData)

    # CFArray functions
    ctypedef struct __CFArray
    ctypedef __CFArray* CFArrayRef
    cdef CFIndex CFArrayGetCount(CFArrayRef theArray)
    cdef void* CFArrayGetValueAtIndex(CFArrayRef theArray, CFIndex idx)

    # CFDictionary functions
    ctypedef struct __CFDictionary
    ctypedef __CFDictionary* CFDictionaryRef
    cdef void* CFDictionaryGetValue(CFDictionaryRef theDict, const void* key)

    # CFNumber functions
    ctypedef struct __CFNumber
    ctypedef __CFNumber* CFNumberRef
    ctypedef enum CFNumberType:
        kCFNumberSInt8Type = 1
        kCFNumberSInt16Type = 2
        kCFNumberSInt32Type = 3
        kCFNumberSInt64Type = 4
        kCFNumberFloat32Type = 5
        kCFNumberFloat64Type = 6
        kCFNumberCharType = 7
        kCFNumberShortType = 8
        kCFNumberIntType = 9
        kCFNumberLongType = 10
        kCFNumberLongLongType = 11
        kCFNumberFloatType = 12
        kCFNumberDoubleType = 13
        kCFNumberCFIndexType = 14
    cdef Boolean CFNumberGetValue(CFNumberRef number, CFNumberType theType, void* valuePtr)


"""pytest test suite for AudioServices functionality."""

import logging
import os
import pytest
import tempfile
import coremusic as cm
import coremusic.capi as capi

logger = logging.getLogger(__name__)


class TestAudioServicesConstants:
    """Test AudioServices constants access"""

    def test_audio_services_error_constants(self):
        """Test AudioServices error constants"""
        assert capi.get_audio_services_no_error() == 0
        assert capi.get_audio_services_unsupported_property_error() == 1886681407
        assert capi.get_audio_services_bad_property_size_error() == 561211770
        assert capi.get_audio_services_bad_specifier_size_error() == 561213539
        assert capi.get_audio_services_system_sound_unspecified_error() == -1500
        assert capi.get_audio_services_system_sound_client_timed_out_error() == -1501
        assert (
            capi.get_audio_services_system_sound_exceeded_maximum_duration_error()
            == -1502
        )

    def test_system_sound_id_constants(self):
        """Test SystemSoundID constants"""
        assert capi.get_system_sound_id_user_preferred_alert() == 4096
        assert capi.get_system_sound_id_flash_screen() == 4094
        assert capi.get_system_sound_id_vibrate() == 4095
        assert capi.get_user_preferred_alert() == 4096

    def test_audio_services_property_constants(self):
        """Test AudioServices property constants"""
        assert capi.get_audio_services_property_is_ui_sound() == 1769174377
        assert (
            capi.get_audio_services_property_complete_playback_if_app_dies()
            == 1768318057
        )


class TestAudioServicesBasicOperations:
    """Test basic AudioServices operations"""

    @pytest.fixture
    def test_audio_file_path(self, amen_wav_path):
        """Path to test audio file"""
        return amen_wav_path

    def test_create_and_dispose_system_sound_id(self, test_audio_file_path):
        """Test creating and disposing a SystemSoundID"""
        sound_id = capi.audio_services_create_system_sound_id(test_audio_file_path)
        assert isinstance(sound_id, int)
        assert sound_id != 0
        result = capi.audio_services_dispose_system_sound_id(sound_id)
        assert result == 0

    def test_create_system_sound_id_invalid_path(self):
        """Test creating SystemSoundID with invalid path"""
        with pytest.raises(RuntimeError):
            capi.audio_services_create_system_sound_id("/path/that/does/not/exist.wav")

    def test_predefined_system_sounds(self):
        """Test playing predefined system sounds"""
        try:
            alert_id = capi.get_system_sound_id_user_preferred_alert()
            capi.audio_services_play_alert_sound(alert_id)
        except Exception:
            pass
        try:
            vibrate_id = capi.get_system_sound_id_vibrate()
            capi.audio_services_play_system_sound(vibrate_id)
        except Exception:
            pass


class TestAudioServicesPlayback:
    """Test AudioServices playback functionality"""

    @pytest.fixture
    def test_audio_file_path(self, amen_wav_path):
        """Path to test audio file"""
        return amen_wav_path

    @pytest.fixture
    def system_sound_id(self, test_audio_file_path):
        """Create a system sound ID for testing"""
        sound_id = capi.audio_services_create_system_sound_id(test_audio_file_path)
        yield sound_id
        capi.audio_services_dispose_system_sound_id(sound_id)

    def test_play_system_sound(self, system_sound_id):
        """Test playing a system sound"""
        capi.audio_services_play_system_sound(system_sound_id)

    def test_play_alert_sound(self, system_sound_id):
        """Test playing an alert sound"""
        capi.audio_services_play_alert_sound(system_sound_id)

    def test_dispose_invalid_sound_id(self):
        """Test disposing an invalid sound ID"""
        try:
            result = capi.audio_services_dispose_system_sound_id(999999)
            assert isinstance(result, int)
        except RuntimeError:
            pass


class TestAudioServicesProperties:
    """Test AudioServices property functionality"""

    @pytest.fixture
    def test_audio_file_path(self, amen_wav_path):
        """Path to test audio file"""
        return amen_wav_path

    @pytest.fixture
    def system_sound_id(self, test_audio_file_path):
        """Create a system sound ID for testing"""
        sound_id = capi.audio_services_create_system_sound_id(test_audio_file_path)
        yield sound_id
        capi.audio_services_dispose_system_sound_id(sound_id)

    def test_get_is_ui_sound_property(self, system_sound_id):
        """Test getting the IsUISound property"""
        try:
            prop_id = capi.get_audio_services_property_is_ui_sound()
            value = capi.audio_services_get_property(prop_id, system_sound_id)
            assert isinstance(value, int)
            assert value in [0, 1]
        except RuntimeError as e:
            print(f"IsUISound property not supported: {e}")

    def test_set_is_ui_sound_property(self, system_sound_id):
        """Test setting the IsUISound property"""
        try:
            prop_id = capi.get_audio_services_property_is_ui_sound()
            result = capi.audio_services_set_property(prop_id, 1, system_sound_id)
            assert result == 0
            value = capi.audio_services_get_property(prop_id, system_sound_id)
            assert value == 1
            result = capi.audio_services_set_property(prop_id, 0, system_sound_id)
            assert result == 0
            value = capi.audio_services_get_property(prop_id, system_sound_id)
            assert value == 0
        except RuntimeError as e:
            print(f"IsUISound property setting not supported: {e}")

    def test_get_complete_playback_if_app_dies_property(self, system_sound_id):
        """Test getting the CompletePlaybackIfAppDies property"""
        try:
            prop_id = capi.get_audio_services_property_complete_playback_if_app_dies()
            value = capi.audio_services_get_property(prop_id, system_sound_id)
            assert isinstance(value, int)
            assert value in [0, 1]
        except RuntimeError as e:
            print(f"CompletePlaybackIfAppDies property not supported: {e}")

    def test_set_complete_playback_if_app_dies_property(self, system_sound_id):
        """Test setting the CompletePlaybackIfAppDies property"""
        try:
            prop_id = capi.get_audio_services_property_complete_playback_if_app_dies()
            result = capi.audio_services_set_property(prop_id, 1, system_sound_id)
            assert result == 0
            value = capi.audio_services_get_property(prop_id, system_sound_id)
            assert value == 1
        except RuntimeError as e:
            print(f"CompletePlaybackIfAppDies property setting not supported: {e}")

    def test_invalid_property_access(self, system_sound_id):
        """Test accessing invalid properties"""
        with pytest.raises(RuntimeError):
            capi.audio_services_get_property(999999, system_sound_id)
        with pytest.raises(RuntimeError):
            capi.audio_services_set_property(999999, 1, system_sound_id)


class TestAudioServicesErrorHandling:
    """Test AudioServices error handling"""

    def test_invalid_file_path_handling(self):
        """Test handling of invalid file paths"""
        invalid_paths = [
            "/this/path/does/not/exist.wav",
            "",
            "/dev/null",
            "/tmp/nonexistent.mp3",
        ]
        for path in invalid_paths:
            with pytest.raises((RuntimeError, ValueError)):
                capi.audio_services_create_system_sound_id(path)

    def test_property_type_validation(self, amen_wav_path):
        """Test property data type validation"""
        wav_path = amen_wav_path
        sound_id = capi.audio_services_create_system_sound_id(wav_path)
        try:
            prop_id = capi.get_audio_services_property_is_ui_sound()
            with pytest.raises(TypeError):
                capi.audio_services_set_property(prop_id, "invalid_string", sound_id)
            with pytest.raises(TypeError):
                capi.audio_services_set_property(prop_id, [1, 2, 3], sound_id)
        finally:
            capi.audio_services_dispose_system_sound_id(sound_id)


class TestAudioServicesResourceManagement:
    """Test AudioServices resource management"""

    def test_multiple_sound_creation_and_disposal(self, amen_wav_path):
        """Test creating and disposing multiple system sounds"""
        wav_path = amen_wav_path
        sound_ids = []
        try:
            for i in range(5):
                sound_id = capi.audio_services_create_system_sound_id(wav_path)
                sound_ids.append(sound_id)
                assert isinstance(sound_id, int)
                assert sound_id != 0
            assert len(set(sound_ids)) == len(sound_ids)
        finally:
            for i, sound_id in enumerate(sound_ids):
                try:
                    capi.audio_services_dispose_system_sound_id(sound_id)
                except Exception as e:
                    logger.warning(f"Cleanup failed for sound_id {i}: {e}")

    def test_sound_id_lifecycle(self, amen_wav_path):
        """Test the complete lifecycle of a SystemSoundID"""
        wav_path = amen_wav_path
        sound_id = capi.audio_services_create_system_sound_id(wav_path)
        assert isinstance(sound_id, int)
        capi.audio_services_play_system_sound(sound_id)
        try:
            prop_id = capi.get_audio_services_property_is_ui_sound()
            capi.audio_services_set_property(prop_id, 0, sound_id)
        except RuntimeError:
            pass
        capi.audio_services_play_system_sound(sound_id)
        result = capi.audio_services_dispose_system_sound_id(sound_id)
        assert result == 0


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#!/usr/bin/env python3
"""Music theory primitives and utilities.

This module provides foundational music theory constructs:
- Note representation with octave and accidentals
- Intervals (semitone distances with quality)
- Scales (major, minor, modes, exotic scales)
- Chords (triads, sevenths, extended, altered)
- Key signatures and circle of fifths

All classes are designed to work seamlessly with MIDI note numbers
(0-127) while providing human-readable interfaces.

Example:
    >>> note = Note('C', 4)  # Middle C
    >>> note.midi  # 60
    >>> note.transpose(Interval.PERFECT_FIFTH)  # G4
    >>>
    >>> scale = Scale(Note('C', 4), ScaleType.MAJOR)
    >>> [n.name for n in scale.get_notes()]  # ['C', 'D', 'E', 'F', 'G', 'A', 'B']
    >>>
    >>> chord = Chord(Note('C', 4), ChordType.MAJOR_7)
    >>> chord.get_midi_notes()  # [60, 64, 67, 71]
"""

import re
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Dict, Iterator, List, Optional, Tuple, Union

# ============================================================================
# Constants
# ============================================================================

# Note names in chromatic order
NOTE_NAMES = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']

# Enharmonic equivalents (uppercase keys for lookup after .upper())
ENHARMONIC_MAP = {
    'DB': 'C#', 'EB': 'D#', 'FB': 'E', 'GB': 'F#', 'AB': 'G#', 'BB': 'A#', 'CB': 'B',
    'B#': 'C', 'E#': 'F', 'C##': 'D', 'D##': 'E', 'F##': 'G', 'G##': 'A', 'A##': 'B',
    'DBB': 'C', 'EBB': 'D', 'FBB': 'E', 'GBB': 'F', 'ABB': 'G', 'BBB': 'A', 'CBB': 'B',
}

# Flat note names for display
FLAT_NAMES = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B']

# Key signatures: key -> (sharps/flats list, is_minor)
KEY_SIGNATURES: Dict[str, Tuple[List[str], bool]] = {
    # Major keys
    'C': ([], False),
    'G': (['F#'], False),
    'D': (['F#', 'C#'], False),
    'A': (['F#', 'C#', 'G#'], False),
    'E': (['F#', 'C#', 'G#', 'D#'], False),
    'B': (['F#', 'C#', 'G#', 'D#', 'A#'], False),
    'F#': (['F#', 'C#', 'G#', 'D#', 'A#', 'E#'], False),
    'C#': (['F#', 'C#', 'G#', 'D#', 'A#', 'E#', 'B#'], False),
    'F': (['Bb'], False),
    'Bb': (['Bb', 'Eb'], False),
    'Eb': (['Bb', 'Eb', 'Ab'], False),
    'Ab': (['Bb', 'Eb', 'Ab', 'Db'], False),
    'Db': (['Bb', 'Eb', 'Ab', 'Db', 'Gb'], False),
    'Gb': (['Bb', 'Eb', 'Ab', 'Db', 'Gb', 'Cb'], False),
    'Cb': (['Bb', 'Eb', 'Ab', 'Db', 'Gb', 'Cb', 'Fb'], False),
    # Minor keys (relative minors)
    'Am': ([], True),
    'Em': (['F#'], True),
    'Bm': (['F#', 'C#'], True),
    'F#m': (['F#', 'C#', 'G#'], True),
    'C#m': (['F#', 'C#', 'G#', 'D#'], True),
    'G#m': (['F#', 'C#', 'G#', 'D#', 'A#'], True),
    'D#m': (['F#', 'C#', 'G#', 'D#', 'A#', 'E#'], True),
    'A#m': (['F#', 'C#', 'G#', 'D#', 'A#', 'E#', 'B#'], True),
    'Dm': (['Bb'], True),
    'Gm': (['Bb', 'Eb'], True),
    'Cm': (['Bb', 'Eb', 'Ab'], True),
    'Fm': (['Bb', 'Eb', 'Ab', 'Db'], True),
    'Bbm': (['Bb', 'Eb', 'Ab', 'Db', 'Gb'], True),
    'Ebm': (['Bb', 'Eb', 'Ab', 'Db', 'Gb', 'Cb'], True),
    'Abm': (['Bb', 'Eb', 'Ab', 'Db', 'Gb', 'Cb', 'Fb'], True),
}

# Circle of fifths (clockwise from C)
CIRCLE_OF_FIFTHS = ['C', 'G', 'D', 'A', 'E', 'B', 'F#', 'Db', 'Ab', 'Eb', 'Bb', 'F']


# ============================================================================
# Utility Functions
# ============================================================================


def note_name_to_midi(name: str, octave: int = 4) -> int:
    """Convert note name to MIDI note number.

    Args:
        name: Note name (e.g., 'C', 'C#', 'Db', 'C4')
        octave: Octave number (default 4, middle C octave)

    Returns:
        MIDI note number (0-127)

    Raises:
        ValueError: If note name is invalid

    Example:
        >>> note_name_to_midi('C', 4)
        60
        >>> note_name_to_midi('A', 4)
        69
        >>> note_name_to_midi('C#4')
        61
    """
    # Parse octave from name if present (e.g., "C4", "F#3")
    match = re.match(r'^([A-Ga-g][#b]*)(-?\d+)?$', name)
    if not match:
        raise ValueError(f"Invalid note name: {name}")

    note_part = match.group(1).upper()
    if match.group(2) is not None:
        octave = int(match.group(2))

    # Handle enharmonic equivalents
    if note_part in ENHARMONIC_MAP:
        note_part = ENHARMONIC_MAP[note_part]

    # Find pitch class
    try:
        pitch_class = NOTE_NAMES.index(note_part)
    except ValueError:
        raise ValueError(f"Unknown note name: {note_part}")

    midi = (octave + 1) * 12 + pitch_class
    if not 0 <= midi <= 127:
        raise ValueError(f"MIDI note {midi} out of range (0-127)")

    return midi


def midi_to_note_name(midi: int, use_flats: bool = False) -> str:
    """Convert MIDI note number to note name with octave.

    Args:
        midi: MIDI note number (0-127)
        use_flats: If True, use flat names (Db) instead of sharps (C#)

    Returns:
        Note name with octave (e.g., 'C4', 'F#3')

    Example:
        >>> midi_to_note_name(60)
        'C4'
        >>> midi_to_note_name(61)
        'C#4'
        >>> midi_to_note_name(61, use_flats=True)
        'Db4'
    """
    if not 0 <= midi <= 127:
        raise ValueError(f"MIDI note {midi} out of range (0-127)")

    octave = (midi // 12) - 1
    pitch_class = midi % 12

    names = FLAT_NAMES if use_flats else NOTE_NAMES
    return f"{names[pitch_class]}{octave}"


# ============================================================================
# Interval Types
# ============================================================================


class IntervalQuality(Enum):
    """Interval quality (perfect, major, minor, augmented, diminished)."""
    PERFECT = "P"
    MAJOR = "M"
    MINOR = "m"
    AUGMENTED = "A"
    DIMINISHED = "d"


@dataclass(frozen=True)
class Interval:
    """Musical interval (distance between two notes).

    Attributes:
        semitones: Number of semitones
        quality: Interval quality
        number: Interval number (1=unison, 2=second, etc.)
        name: Human-readable name

    Example:
        >>> Interval.PERFECT_FIFTH
        Interval(semitones=7, name='Perfect Fifth')
        >>> Interval.from_semitones(7)
        Interval.PERFECT_FIFTH
    """
    semitones: int
    quality: IntervalQuality
    number: int
    name: str

    # Class-level interval constants (assigned below class definition)
    UNISON: ClassVar["Interval"]
    MINOR_SECOND: ClassVar["Interval"]
    MAJOR_SECOND: ClassVar["Interval"]
    MINOR_THIRD: ClassVar["Interval"]
    MAJOR_THIRD: ClassVar["Interval"]
    PERFECT_FOURTH: ClassVar["Interval"]
    TRITONE: ClassVar["Interval"]
    PERFECT_FIFTH: ClassVar["Interval"]
    MINOR_SIXTH: ClassVar["Interval"]
    MAJOR_SIXTH: ClassVar["Interval"]
    MINOR_SEVENTH: ClassVar["Interval"]
    MAJOR_SEVENTH: ClassVar["Interval"]
    OCTAVE: ClassVar["Interval"]
    MINOR_NINTH: ClassVar["Interval"]
    MAJOR_NINTH: ClassVar["Interval"]
    MINOR_TENTH: ClassVar["Interval"]
    MAJOR_TENTH: ClassVar["Interval"]
    PERFECT_ELEVENTH: ClassVar["Interval"]
    AUGMENTED_ELEVENTH: ClassVar["Interval"]
    PERFECT_TWELFTH: ClassVar["Interval"]
    MINOR_THIRTEENTH: ClassVar["Interval"]
    MAJOR_THIRTEENTH: ClassVar["Interval"]

    def __repr__(self) -> str:
        return f"Interval({self.semitones}, {self.name!r})"

    @classmethod
    def from_semitones(cls, semitones: int) -> 'Interval':
        """Create interval from semitone count.

        Args:
            semitones: Number of semitones (0-24 typically)

        Returns:
            Matching Interval constant
        """
        semitones = semitones % 12  # Normalize to one octave
        for interval in INTERVALS:
            if interval.semitones == semitones:
                return interval
        # Default to chromatic interval
        return cls(semitones, IntervalQuality.AUGMENTED, semitones, f"{semitones} semitones")

    @classmethod
    def between(cls, note1: 'Note', note2: 'Note') -> 'Interval':
        """Calculate interval between two notes.

        Args:
            note1: First note
            note2: Second note

        Returns:
            Interval between the notes
        """
        semitones = abs(note2.midi - note1.midi)
        return cls.from_semitones(semitones)


# Standard intervals
Interval.UNISON = Interval(0, IntervalQuality.PERFECT, 1, "Unison")
Interval.MINOR_SECOND = Interval(1, IntervalQuality.MINOR, 2, "Minor Second")
Interval.MAJOR_SECOND = Interval(2, IntervalQuality.MAJOR, 2, "Major Second")
Interval.MINOR_THIRD = Interval(3, IntervalQuality.MINOR, 3, "Minor Third")
Interval.MAJOR_THIRD = Interval(4, IntervalQuality.MAJOR, 3, "Major Third")
Interval.PERFECT_FOURTH = Interval(5, IntervalQuality.PERFECT, 4, "Perfect Fourth")
Interval.TRITONE = Interval(6, IntervalQuality.AUGMENTED, 4, "Tritone")
Interval.PERFECT_FIFTH = Interval(7, IntervalQuality.PERFECT, 5, "Perfect Fifth")
Interval.MINOR_SIXTH = Interval(8, IntervalQuality.MINOR, 6, "Minor Sixth")
Interval.MAJOR_SIXTH = Interval(9, IntervalQuality.MAJOR, 6, "Major Sixth")
Interval.MINOR_SEVENTH = Interval(10, IntervalQuality.MINOR, 7, "Minor Seventh")
Interval.MAJOR_SEVENTH = Interval(11, IntervalQuality.MAJOR, 7, "Major Seventh")
Interval.OCTAVE = Interval(12, IntervalQuality.PERFECT, 8, "Octave")

# Compound intervals
Interval.MINOR_NINTH = Interval(13, IntervalQuality.MINOR, 9, "Minor Ninth")
Interval.MAJOR_NINTH = Interval(14, IntervalQuality.MAJOR, 9, "Major Ninth")
Interval.MINOR_TENTH = Interval(15, IntervalQuality.MINOR, 10, "Minor Tenth")
Interval.MAJOR_TENTH = Interval(16, IntervalQuality.MAJOR, 10, "Major Tenth")
Interval.PERFECT_ELEVENTH = Interval(17, IntervalQuality.PERFECT, 11, "Perfect Eleventh")
Interval.AUGMENTED_ELEVENTH = Interval(18, IntervalQuality.AUGMENTED, 11, "Augmented Eleventh")
Interval.PERFECT_TWELFTH = Interval(19, IntervalQuality.PERFECT, 12, "Perfect Twelfth")
Interval.MINOR_THIRTEENTH = Interval(20, IntervalQuality.MINOR, 13, "Minor Thirteenth")
Interval.MAJOR_THIRTEENTH = Interval(21, IntervalQuality.MAJOR, 13, "Major Thirteenth")

# List of standard intervals
INTERVALS = [
    Interval.UNISON, Interval.MINOR_SECOND, Interval.MAJOR_SECOND,
    Interval.MINOR_THIRD, Interval.MAJOR_THIRD, Interval.PERFECT_FOURTH,
    Interval.TRITONE, Interval.PERFECT_FIFTH, Interval.MINOR_SIXTH,
    Interval.MAJOR_SIXTH, Interval.MINOR_SEVENTH, Interval.MAJOR_SEVENTH,
    Interval.OCTAVE,
]


# ============================================================================
# Note Class
# ============================================================================


@dataclass
class Note:
    """Musical note with pitch class and octave.

    Attributes:
        name: Note name without octave (e.g., 'C', 'F#', 'Bb')
        octave: Octave number (-1 to 9 for MIDI range)
        velocity: Default velocity (0-127)

    Example:
        >>> note = Note('C', 4)  # Middle C
        >>> note.midi
        60
        >>> note.frequency
        261.63
        >>> note.transpose(7)  # G4
        Note('G', 4)
    """
    name: str
    octave: int = 4
    velocity: int = 100

    def __post_init__(self):
        # Normalize note name
        self.name = self.name.upper()
        if self.name in ENHARMONIC_MAP:
            self.name = ENHARMONIC_MAP[self.name]

        # Validate
        if self.name not in NOTE_NAMES:
            raise ValueError(f"Invalid note name: {self.name}")
        if not -1 <= self.octave <= 9:
            raise ValueError(f"Octave {self.octave} out of range (-1 to 9)")
        if not 0 <= self.velocity <= 127:
            raise ValueError(f"Velocity {self.velocity} out of range (0-127)")

    @property
    def midi(self) -> int:
        """MIDI note number (0-127)."""
        return note_name_to_midi(self.name, self.octave)

    @property
    def pitch_class(self) -> int:
        """Pitch class (0-11, where C=0)."""
        return NOTE_NAMES.index(self.name)

    @property
    def frequency(self) -> float:
        """Frequency in Hz (A4 = 440 Hz standard tuning)."""
        return float(440.0 * (2.0 ** ((self.midi - 69) / 12.0)))

    @classmethod
    def from_midi(cls, midi: int, velocity: int = 100, use_flats: bool = False) -> 'Note':
        """Create Note from MIDI number.

        Args:
            midi: MIDI note number (0-127)
            velocity: Note velocity (0-127)
            use_flats: Use flat names instead of sharps

        Returns:
            Note instance
        """
        if not 0 <= midi <= 127:
            raise ValueError(f"MIDI note {midi} out of range (0-127)")

        octave = (midi // 12) - 1
        pitch_class = midi % 12
        names = FLAT_NAMES if use_flats else NOTE_NAMES
        return cls(names[pitch_class], octave, velocity)

    def transpose(self, semitones: Union[int, Interval]) -> 'Note':
        """Transpose note by semitones or interval.

        Args:
            semitones: Number of semitones or Interval

        Returns:
            New transposed Note
        """
        if isinstance(semitones, Interval):
            semitones = semitones.semitones

        new_midi = self.midi + semitones
        if not 0 <= new_midi <= 127:
            raise ValueError(f"Transposed MIDI note {new_midi} out of range")

        return Note.from_midi(new_midi, self.velocity)

    def interval_to(self, other: 'Note') -> Interval:
        """Get interval to another note.

        Args:
            other: Target note

        Returns:
            Interval between this note and other
        """
        return Interval.between(self, other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Note):
            return self.midi == other.midi
        return False

    def __hash__(self) -> int:
        return hash(self.midi)

    def __lt__(self, other: 'Note') -> bool:
        return self.midi < other.midi

    def __repr__(self) -> str:
        return f"Note({self.name!r}, {self.octave})"

    def __str__(self) -> str:
        return f"{self.name}{self.octave}"


# ============================================================================
# Scale Types and Class
# ============================================================================


class ScaleType(Enum):
    """Common scale types with interval patterns (semitones from root)."""
    # Diatonic scales
    MAJOR = (0, 2, 4, 5, 7, 9, 11)
    NATURAL_MINOR = (0, 2, 3, 5, 7, 8, 10)
    HARMONIC_MINOR = (0, 2, 3, 5, 7, 8, 11)
    MELODIC_MINOR = (0, 2, 3, 5, 7, 9, 11)

    # Modes
    IONIAN = (0, 2, 4, 5, 7, 9, 11)  # Same as major
    DORIAN = (0, 2, 3, 5, 7, 9, 10)
    PHRYGIAN = (0, 1, 3, 5, 7, 8, 10)
    LYDIAN = (0, 2, 4, 6, 7, 9, 11)
    MIXOLYDIAN = (0, 2, 4, 5, 7, 9, 10)
    AEOLIAN = (0, 2, 3, 5, 7, 8, 10)  # Same as natural minor
    LOCRIAN = (0, 1, 3, 5, 6, 8, 10)

    # Pentatonic scales
    MAJOR_PENTATONIC = (0, 2, 4, 7, 9)
    MINOR_PENTATONIC = (0, 3, 5, 7, 10)

    # Blues scales
    BLUES = (0, 3, 5, 6, 7, 10)
    MAJOR_BLUES = (0, 2, 3, 4, 7, 9)

    # Symmetric scales
    WHOLE_TONE = (0, 2, 4, 6, 8, 10)
    DIMINISHED = (0, 2, 3, 5, 6, 8, 9, 11)  # Half-whole
    DIMINISHED_WHOLE_HALF = (0, 1, 3, 4, 6, 7, 9, 10)  # Whole-half
    CHROMATIC = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)

    # Exotic scales
    HUNGARIAN_MINOR = (0, 2, 3, 6, 7, 8, 11)
    PHRYGIAN_DOMINANT = (0, 1, 4, 5, 7, 8, 10)  # Spanish/Jewish
    DOUBLE_HARMONIC = (0, 1, 4, 5, 7, 8, 11)  # Byzantine
    HIRAJOSHI = (0, 2, 3, 7, 8)  # Japanese
    IN_SEN = (0, 1, 5, 7, 10)  # Japanese
    IWATO = (0, 1, 5, 6, 10)  # Japanese
    ARABIAN = (0, 2, 4, 5, 6, 8, 10)
    PERSIAN = (0, 1, 4, 5, 6, 8, 11)
    BALINESE = (0, 1, 3, 7, 8)


class Mode(Enum):
    """Church modes (relative to major scale)."""
    IONIAN = 0      # I - Major
    DORIAN = 1      # II
    PHRYGIAN = 2    # III
    LYDIAN = 3      # IV
    MIXOLYDIAN = 4  # V
    AEOLIAN = 5     # VI - Natural minor
    LOCRIAN = 6     # VII


@dataclass
class Scale:
    """Musical scale from a root note.

    Attributes:
        root: Root note of the scale
        scale_type: Type of scale (intervals pattern)
        octaves: Number of octaves to generate

    Example:
        >>> scale = Scale(Note('C', 4), ScaleType.MAJOR)
        >>> notes = scale.get_notes()
        >>> [str(n) for n in notes]
        ['C4', 'D4', 'E4', 'F4', 'G4', 'A4', 'B4']
        >>>
        >>> scale.contains(Note('E', 4))
        True
        >>> scale.degree(3)  # Third degree
        Note('E', 4)
    """
    root: Note
    scale_type: ScaleType
    octaves: int = 1

    @property
    def intervals(self) -> Tuple[int, ...]:
        """Semitone intervals from root."""
        result: Tuple[int, ...] = self.scale_type.value
        return result

    def get_notes(self, octaves: Optional[int] = None) -> List[Note]:
        """Get all notes in the scale.

        Args:
            octaves: Number of octaves (default: self.octaves)

        Returns:
            List of Notes in the scale
        """
        if octaves is None:
            octaves = self.octaves

        notes = []
        for octave_offset in range(octaves):
            for interval in self.intervals:
                midi = self.root.midi + interval + (octave_offset * 12)
                if 0 <= midi <= 127:
                    notes.append(Note.from_midi(midi, self.root.velocity))

        return notes

    def get_midi_notes(self, octaves: Optional[int] = None) -> List[int]:
        """Get MIDI note numbers in the scale.

        Args:
            octaves: Number of octaves

        Returns:
            List of MIDI note numbers
        """
        return [n.midi for n in self.get_notes(octaves)]

    def degree(self, degree: int) -> Note:
        """Get note at specific scale degree.

        Args:
            degree: Scale degree (1-based, 1 = root)

        Returns:
            Note at that degree
        """
        if degree < 1:
            raise ValueError(f"Degree must be >= 1, got {degree}")

        # Handle degrees beyond one octave
        octave_offset = (degree - 1) // len(self.intervals)
        degree_in_octave = (degree - 1) % len(self.intervals)

        interval = self.intervals[degree_in_octave]
        midi = self.root.midi + interval + (octave_offset * 12)

        if not 0 <= midi <= 127:
            raise ValueError(f"Degree {degree} results in MIDI {midi} out of range")

        return Note.from_midi(midi, self.root.velocity)

    def contains(self, note: Note) -> bool:
        """Check if note is in the scale.

        Args:
            note: Note to check

        Returns:
            True if note's pitch class is in the scale
        """
        root_pc = self.root.pitch_class
        note_pc = note.pitch_class
        relative_pc = (note_pc - root_pc) % 12
        return relative_pc in self.intervals

    def harmonize(self, degree: int, chord_type: Optional['ChordType'] = None) -> 'Chord':
        """Build chord on scale degree.

        Args:
            degree: Scale degree (1-based)
            chord_type: Chord type, or None for diatonic triad

        Returns:
            Chord built on that degree
        """
        root = self.degree(degree)

        if chord_type is not None:
            return Chord(root, chord_type)

        # Build diatonic triad
        third = self.degree(degree + 2)
        fifth = self.degree(degree + 4)

        third_interval = (third.midi - root.midi) % 12
        fifth_interval = (fifth.midi - root.midi) % 12

        # Determine chord type from intervals
        if third_interval == 4 and fifth_interval == 7:
            return Chord(root, ChordType.MAJOR)
        elif third_interval == 3 and fifth_interval == 7:
            return Chord(root, ChordType.MINOR)
        elif third_interval == 3 and fifth_interval == 6:
            return Chord(root, ChordType.DIMINISHED)
        elif third_interval == 4 and fifth_interval == 8:
            return Chord(root, ChordType.AUGMENTED)
        else:
            return Chord(root, ChordType.MAJOR)  # Default

    def parallel(self, scale_type: ScaleType) -> 'Scale':
        """Get parallel scale (same root, different type).

        Args:
            scale_type: New scale type

        Returns:
            Parallel scale
        """
        return Scale(self.root, scale_type, self.octaves)

    def relative_minor(self) -> 'Scale':
        """Get relative minor scale.

        Returns:
            Relative minor scale (6th degree becomes root)
        """
        minor_root = self.degree(6)
        return Scale(minor_root, ScaleType.NATURAL_MINOR, self.octaves)

    def relative_major(self) -> 'Scale':
        """Get relative major scale.

        Returns:
            Relative major scale (3rd degree becomes root)
        """
        major_root = self.degree(3)
        return Scale(major_root, ScaleType.MAJOR, self.octaves)

    def __iter__(self) -> Iterator[Note]:
        """Iterate over scale notes."""
        return iter(self.get_notes())

    def __len__(self) -> int:
        """Number of notes in one octave of the scale."""
        return len(self.intervals)

    def __repr__(self) -> str:
        return f"Scale({self.root}, {self.scale_type.name})"


# ============================================================================
# Chord Types and Class
# ============================================================================


class ChordType(Enum):
    """Common chord types with interval patterns (semitones from root)."""
    # Triads
    MAJOR = (0, 4, 7)
    MINOR = (0, 3, 7)
    DIMINISHED = (0, 3, 6)
    AUGMENTED = (0, 4, 8)
    SUS2 = (0, 2, 7)
    SUS4 = (0, 5, 7)

    # Seventh chords
    MAJOR_7 = (0, 4, 7, 11)
    MINOR_7 = (0, 3, 7, 10)
    DOMINANT_7 = (0, 4, 7, 10)
    DIMINISHED_7 = (0, 3, 6, 9)
    HALF_DIMINISHED_7 = (0, 3, 6, 10)  # Minor 7 flat 5
    MINOR_MAJOR_7 = (0, 3, 7, 11)
    AUGMENTED_7 = (0, 4, 8, 10)
    AUGMENTED_MAJOR_7 = (0, 4, 8, 11)

    # Sixth chords
    MAJOR_6 = (0, 4, 7, 9)
    MINOR_6 = (0, 3, 7, 9)

    # Add chords
    ADD9 = (0, 4, 7, 14)
    ADD11 = (0, 4, 7, 17)
    MINOR_ADD9 = (0, 3, 7, 14)

    # Extended chords
    MAJOR_9 = (0, 4, 7, 11, 14)
    MINOR_9 = (0, 3, 7, 10, 14)
    DOMINANT_9 = (0, 4, 7, 10, 14)
    MAJOR_11 = (0, 4, 7, 11, 14, 17)
    MINOR_11 = (0, 3, 7, 10, 14, 17)
    DOMINANT_11 = (0, 4, 7, 10, 14, 17)
    MAJOR_13 = (0, 4, 7, 11, 14, 17, 21)
    MINOR_13 = (0, 3, 7, 10, 14, 17, 21)
    DOMINANT_13 = (0, 4, 7, 10, 14, 17, 21)

    # Altered chords
    DOMINANT_7_FLAT_5 = (0, 4, 6, 10)
    DOMINANT_7_SHARP_5 = (0, 4, 8, 10)
    DOMINANT_7_FLAT_9 = (0, 4, 7, 10, 13)
    DOMINANT_7_SHARP_9 = (0, 4, 7, 10, 15)
    DOMINANT_7_FLAT_5_FLAT_9 = (0, 4, 6, 10, 13)
    DOMINANT_7_SHARP_5_SHARP_9 = (0, 4, 8, 10, 15)

    # Power chord
    POWER = (0, 7)
    POWER_OCTAVE = (0, 7, 12)


# Chord symbol mappings for parsing
CHORD_SYMBOL_MAP = {
    '': ChordType.MAJOR,
    'maj': ChordType.MAJOR,
    'M': ChordType.MAJOR,
    'm': ChordType.MINOR,
    'min': ChordType.MINOR,
    '-': ChordType.MINOR,
    'dim': ChordType.DIMINISHED,
    'o': ChordType.DIMINISHED,
    'aug': ChordType.AUGMENTED,
    '+': ChordType.AUGMENTED,
    'sus2': ChordType.SUS2,
    'sus4': ChordType.SUS4,
    'sus': ChordType.SUS4,
    '7': ChordType.DOMINANT_7,
    'maj7': ChordType.MAJOR_7,
    'M7': ChordType.MAJOR_7,
    'm7': ChordType.MINOR_7,
    'min7': ChordType.MINOR_7,
    '-7': ChordType.MINOR_7,
    'dim7': ChordType.DIMINISHED_7,
    'o7': ChordType.DIMINISHED_7,
    'm7b5': ChordType.HALF_DIMINISHED_7,
    'mM7': ChordType.MINOR_MAJOR_7,
    '6': ChordType.MAJOR_6,
    'm6': ChordType.MINOR_6,
    '9': ChordType.DOMINANT_9,
    'maj9': ChordType.MAJOR_9,
    'm9': ChordType.MINOR_9,
    '11': ChordType.DOMINANT_11,
    '13': ChordType.DOMINANT_13,
    'add9': ChordType.ADD9,
    '5': ChordType.POWER,
}


@dataclass
class Chord:
    """Musical chord from a root note.

    Attributes:
        root: Root note of the chord
        chord_type: Type of chord (interval pattern)

    Example:
        >>> chord = Chord(Note('C', 4), ChordType.MAJOR_7)
        >>> chord.get_midi_notes()
        [60, 64, 67, 71]
        >>>
        >>> chord.inversion(1)
        Chord with E as bass note
    """
    root: Note
    chord_type: ChordType

    @property
    def intervals(self) -> Tuple[int, ...]:
        """Semitone intervals from root."""
        result: Tuple[int, ...] = self.chord_type.value
        return result

    def get_notes(self) -> List[Note]:
        """Get all notes in the chord.

        Returns:
            List of Notes
        """
        notes = []
        for interval in self.intervals:
            midi = self.root.midi + interval
            if 0 <= midi <= 127:
                notes.append(Note.from_midi(midi, self.root.velocity))
        return notes

    def get_midi_notes(self) -> List[int]:
        """Get MIDI note numbers in the chord.

        Returns:
            List of MIDI note numbers
        """
        return [n.midi for n in self.get_notes()]

    def inversion(self, n: int) -> 'Chord':
        """Get chord inversion.

        Args:
            n: Inversion number (0=root, 1=first, 2=second, etc.)

        Returns:
            New Chord representing the inversion

        Note:
            Returns a chord with reordered intervals, not a truly new voicing.
            For more complex voicings, use get_voicing().
        """
        notes = self.get_notes()
        if n >= len(notes):
            n = n % len(notes)

        # Rotate notes and adjust octaves
        rotated = notes[n:] + [Note.from_midi(note.midi + 12) for note in notes[:n]]

        # Create new chord from lowest note
        return Chord(rotated[0], self.chord_type)

    def get_voicing(self, voicing: List[int]) -> List[Note]:
        """Get specific voicing of the chord.

        Args:
            voicing: List of interval indices (0=root, 1=third, etc.)
                     Can include negative (below root) or >len (octave up)

        Returns:
            List of Notes in the specified voicing

        Example:
            >>> chord = Chord(Note('C', 4), ChordType.MAJOR_7)
            >>> # Drop-2 voicing: root, 5th, 7th, 3rd (octave up)
            >>> chord.get_voicing([0, 2, 3, 5])  # 5 = 1 + 4 (3rd up octave)
        """
        notes = []
        num_intervals = len(self.intervals)

        for idx in voicing:
            octave_offset = idx // num_intervals
            interval_idx = idx % num_intervals
            interval = self.intervals[interval_idx]
            midi = self.root.midi + interval + (octave_offset * 12)

            if 0 <= midi <= 127:
                notes.append(Note.from_midi(midi, self.root.velocity))

        return notes

    def transpose(self, semitones: int) -> 'Chord':
        """Transpose chord by semitones.

        Args:
            semitones: Number of semitones

        Returns:
            Transposed chord
        """
        new_root = self.root.transpose(semitones)
        return Chord(new_root, self.chord_type)

    @classmethod
    def from_symbol(cls, symbol: str, octave: int = 4) -> 'Chord':
        """Parse chord from symbol string.

        Args:
            symbol: Chord symbol (e.g., 'Cmaj7', 'F#m', 'Bb7')
            octave: Root note octave

        Returns:
            Chord instance

        Example:
            >>> Chord.from_symbol('Cmaj7')
            Chord(Note('C', 4), ChordType.MAJOR_7)
            >>> Chord.from_symbol('F#m')
            Chord(Note('F#', 4), ChordType.MINOR)
        """
        # Parse root note (handles sharps/flats)
        match = re.match(r'^([A-Ga-g][#b]?)(.*?)$', symbol)
        if not match:
            raise ValueError(f"Invalid chord symbol: {symbol}")

        root_name = match.group(1).upper()
        quality = match.group(2)

        # Handle enharmonic
        if root_name in ENHARMONIC_MAP:
            root_name = ENHARMONIC_MAP[root_name]

        root = Note(root_name, octave)

        # Look up chord type
        chord_type = CHORD_SYMBOL_MAP.get(quality)
        if chord_type is None:
            raise ValueError(f"Unknown chord quality: {quality}")

        return cls(root, chord_type)

    @property
    def symbol(self) -> str:
        """Get chord symbol string."""
        # Reverse lookup in symbol map
        for sym, ct in CHORD_SYMBOL_MAP.items():
            if ct == self.chord_type and sym:  # Skip empty string
                return f"{self.root.name}{sym}"
        return f"{self.root.name}"

    def __iter__(self) -> Iterator[Note]:
        """Iterate over chord notes."""
        return iter(self.get_notes())

    def __len__(self) -> int:
        """Number of notes in the chord."""
        return len(self.intervals)

    def __repr__(self) -> str:
        return f"Chord({self.root}, {self.chord_type.name})"

    def __str__(self) -> str:
        return self.symbol


# ============================================================================
# Chord Progression
# ============================================================================


@dataclass
class ChordProgression:
    """A sequence of chords.

    Attributes:
        chords: List of chords
        key: Key signature (root note name)
        scale_type: Scale type for the key

    Example:
        >>> prog = ChordProgression.from_numerals('C', ['I', 'V', 'vi', 'IV'])
        >>> [str(c) for c in prog.chords]
        ['Cmaj', 'Gmaj', 'Am', 'Fmaj']
    """
    chords: List[Chord]
    key: Optional[str] = None
    scale_type: ScaleType = ScaleType.MAJOR

    @classmethod
    def from_numerals(
        cls,
        key: str,
        numerals: List[str],
        octave: int = 4,
        scale_type: ScaleType = ScaleType.MAJOR,
    ) -> 'ChordProgression':
        """Create progression from Roman numeral notation.

        Args:
            key: Root note name (e.g., 'C', 'F#')
            numerals: List of Roman numerals (e.g., ['I', 'IV', 'V', 'I'])
            octave: Base octave
            scale_type: Scale type

        Returns:
            ChordProgression instance

        Example:
            >>> ChordProgression.from_numerals('C', ['I', 'V', 'vi', 'IV'])
        """
        numeral_map = {
            'I': 1, 'II': 2, 'III': 3, 'IV': 4, 'V': 5, 'VI': 6, 'VII': 7,
            'i': 1, 'ii': 2, 'iii': 3, 'iv': 4, 'v': 5, 'vi': 6, 'vii': 7,
        }

        root = Note(key, octave)
        scale = Scale(root, scale_type)
        chords = []

        for numeral in numerals:
            # Check for additional quality modifiers
            base_numeral = numeral.rstrip('o+7')
            is_minor = base_numeral.islower()
            is_dim = 'o' in numeral
            is_aug = '+' in numeral
            has_7th = '7' in numeral

            degree = numeral_map.get(base_numeral.upper())
            if degree is None:
                raise ValueError(f"Invalid numeral: {numeral}")

            chord_root = scale.degree(degree)

            # Determine chord type
            if is_dim:
                chord_type = ChordType.DIMINISHED_7 if has_7th else ChordType.DIMINISHED
            elif is_aug:
                chord_type = ChordType.AUGMENTED
            elif is_minor:
                chord_type = ChordType.MINOR_7 if has_7th else ChordType.MINOR
            else:
                if has_7th:
                    # V7 is dominant, others are major 7
                    chord_type = ChordType.DOMINANT_7 if degree == 5 else ChordType.MAJOR_7
                else:
                    chord_type = ChordType.MAJOR

            chords.append(Chord(chord_root, chord_type))

        return cls(chords, key, scale_type)

    @classmethod
    def from_symbols(cls, symbols: List[str], octave: int = 4) -> 'ChordProgression':
        """Create progression from chord symbols.

        Args:
            symbols: List of chord symbols (e.g., ['Cmaj7', 'Dm7', 'G7'])
            octave: Base octave

        Returns:
            ChordProgression instance
        """
        chords = [Chord.from_symbol(sym, octave) for sym in symbols]
        return cls(chords)

    def transpose(self, semitones: int) -> 'ChordProgression':
        """Transpose entire progression.

        Args:
            semitones: Number of semitones

        Returns:
            Transposed progression
        """
        new_chords = [c.transpose(semitones) for c in self.chords]
        new_key = None
        if self.key:
            root = Note(self.key, 4)
            new_root = root.transpose(semitones)
            new_key = new_root.name
        return ChordProgression(new_chords, new_key, self.scale_type)

    def __iter__(self) -> Iterator[Chord]:
        """Iterate over chords."""
        return iter(self.chords)

    def __len__(self) -> int:
        """Number of chords."""
        return len(self.chords)

    def __repr__(self) -> str:
        symbols = [str(c) for c in self.chords]
        return f"ChordProgression({symbols})"
